from sqlglot import exp, parse_one
from sqlglot.generator import logger as generator_logger
from tests.dialects.test_dialect import Validator


class TestTrino(Validator):
    dialect = "trino"

    def test_trino(self):
        self.validate_identity("REFRESH MATERIALIZED VIEW mynamespace.test_view")
        self.validate_identity("JSON_QUERY(m.properties, 'lax $.area' OMIT QUOTES NULL ON ERROR)")
        self.validate_identity("JSON_EXTRACT(content, json_path)")
        self.validate_identity("JSON_QUERY(content, 'lax $.HY.*')")
        self.validate_identity("JSON_QUERY(content, 'strict $.HY.*' WITH WRAPPER)")
        self.validate_identity("JSON_QUERY(content, 'strict $.HY.*' WITH ARRAY WRAPPER)")
        self.validate_identity("JSON_QUERY(content, 'strict $.HY.*' WITH UNCONDITIONAL WRAPPER)")
        self.validate_identity("JSON_QUERY(content, 'strict $.HY.*' WITHOUT CONDITIONAL WRAPPER)")
        self.validate_identity("JSON_QUERY(description, 'strict $.comment' KEEP QUOTES)")
        self.validate_identity(
            "JSON_QUERY(description, 'strict $.comment' OMIT QUOTES ON SCALAR STRING)"
        )
        self.validate_identity(
            "JSON_QUERY(content, 'strict $.HY.*' WITH UNCONDITIONAL WRAPPER KEEP QUOTES)"
        )
        self.validate_identity(
            "SELECT TIMESTAMP '2012-10-31 01:00 -2'",
            "SELECT CAST('2012-10-31 01:00 -2' AS TIMESTAMP WITH TIME ZONE)",
        )
        self.validate_identity(
            "SELECT TIMESTAMP '2012-10-31 01:00 +2'",
            "SELECT CAST('2012-10-31 01:00 +2' AS TIMESTAMP WITH TIME ZONE)",
        )

        self.validate_all(
            "SELECT FROM_ISO8601_TIMESTAMP_NANOS('2020-05-11T11:15:05.000000000')",
            write={
                "duckdb": "SELECT CAST('2020-05-11T11:15:05.000000000' AS TIMESTAMPTZ)",
                "trino": "SELECT FROM_ISO8601_TIMESTAMP_NANOS('2020-05-11T11:15:05.000000000')",
                "snowflake": "SELECT CAST('2020-05-11T11:15:05.000000000' AS TIMESTAMPTZ)",
                "spark": "SELECT CAST('2020-05-11T11:15:05.000000000' AS TIMESTAMP)",
                "databricks": "SELECT CAST('2020-05-11T11:15:05.000000000' AS TIMESTAMP)",
                "bigquery": "SELECT CAST('2020-05-11T11:15:05.000000000' AS TIMESTAMP)",
            },
        )
        self.validate_all(
            "SELECT TIMESTAMP '2012-10-31 01:00:00 +02:00'",
            write={
                "duckdb": "SELECT CAST('2012-10-31 01:00:00 +02:00' AS TIMESTAMPTZ)",
                "trino": "SELECT CAST('2012-10-31 01:00:00 +02:00' AS TIMESTAMP WITH TIME ZONE)",
            },
        )
        self.validate_all(
            "SELECT FORMAT('%s', 123)",
            write={
                "duckdb": "SELECT FORMAT('{}', 123)",
                "snowflake": "SELECT TO_CHAR(123)",
                "trino": "SELECT FORMAT('%s', 123)",
            },
        )

        self.validate_identity(
            "SELECT * FROM tbl MATCH_RECOGNIZE (PARTITION BY id ORDER BY col MEASURES FIRST(col, 2) AS col1, LAST(col, 2) AS col2 PATTERN (B* A) DEFINE A AS col = 1)"
        )

        self.validate_identity("SELECT VERSION()")

    def test_listagg(self):
        self.validate_identity(
            "SELECT LISTAGG(DISTINCT col, ',') WITHIN GROUP (ORDER BY col ASC) FROM tbl"
        )
        self.validate_identity(
            "SELECT LISTAGG(col, '; ' ON OVERFLOW ERROR) WITHIN GROUP (ORDER BY col ASC) FROM tbl"
        )
        self.validate_identity(
            "SELECT LISTAGG(col, '; ' ON OVERFLOW TRUNCATE WITH COUNT) WITHIN GROUP (ORDER BY col ASC) FROM tbl"
        )
        self.validate_identity(
            "SELECT LISTAGG(col, '; ' ON OVERFLOW TRUNCATE WITHOUT COUNT) WITHIN GROUP (ORDER BY col ASC) FROM tbl"
        )
        self.validate_identity(
            "SELECT LISTAGG(col, '; ' ON OVERFLOW TRUNCATE '...' WITH COUNT) WITHIN GROUP (ORDER BY col ASC) FROM tbl"
        )
        self.validate_identity(
            "SELECT LISTAGG(col, '; ' ON OVERFLOW TRUNCATE '...' WITHOUT COUNT) WITHIN GROUP (ORDER BY col ASC) FROM tbl"
        )
        self.validate_identity(
            "SELECT LISTAGG(col) WITHIN GROUP (ORDER BY col DESC) FROM tbl",
            "SELECT LISTAGG(col, ',') WITHIN GROUP (ORDER BY col DESC) FROM tbl",
        )

    def test_trim(self):
        self.validate_identity("SELECT TRIM('!' FROM '!foo!')")
        self.validate_identity("SELECT TRIM(BOTH '$' FROM '$var$')")
        self.validate_identity("SELECT TRIM(TRAILING 'ER' FROM UPPER('worker'))")
        self.validate_identity(
            "SELECT TRIM(LEADING FROM '  abcd')",
            "SELECT LTRIM('  abcd')",
        )
        self.validate_identity(
            "SELECT TRIM('!foo!', '!')",
            "SELECT TRIM('!' FROM '!foo!')",
        )

    def test_ddl(self):
        self.validate_identity("ALTER TABLE users RENAME TO people")
        self.validate_identity("ALTER TABLE IF EXISTS users RENAME TO people")
        self.validate_identity("ALTER TABLE users ADD COLUMN zip VARCHAR")
        self.validate_identity("ALTER TABLE IF EXISTS users ADD COLUMN IF NOT EXISTS zip VARCHAR")
        self.validate_identity("ALTER TABLE users DROP COLUMN zip")
        self.validate_identity("ALTER TABLE IF EXISTS users DROP COLUMN IF EXISTS zip")
        self.validate_identity("ALTER TABLE users RENAME COLUMN id TO user_id")
        self.validate_identity("ALTER TABLE IF EXISTS users RENAME COLUMN IF EXISTS id TO user_id")
        self.validate_identity("ALTER TABLE users ALTER COLUMN id SET DATA TYPE BIGINT")
        self.validate_identity("ALTER TABLE users ALTER COLUMN id DROP NOT NULL")
        self.validate_identity(
            "ALTER TABLE people SET AUTHORIZATION alice", check_command_warning=True
        )
        self.validate_identity(
            "ALTER TABLE people SET AUTHORIZATION ROLE PUBLIC", check_command_warning=True
        )
        self.validate_identity(
            "ALTER TABLE people SET PROPERTIES x = 'y'", check_command_warning=True
        )
        self.validate_identity(
            "ALTER TABLE people SET PROPERTIES foo = 123, 'foo bar' = 456",
            check_command_warning=True,
        )
        self.validate_identity(
            "ALTER TABLE people SET PROPERTIES x = DEFAULT", check_command_warning=True
        )
        self.validate_identity("ALTER VIEW people RENAME TO users")
        self.validate_identity(
            "ALTER VIEW people SET AUTHORIZATION alice", check_command_warning=True
        )
        self.validate_identity("CREATE SCHEMA foo WITH (LOCATION='s3://bucket/foo')")
        self.validate_identity(
            "CREATE TABLE foo.bar WITH (LOCATION='s3://bucket/foo/bar') AS SELECT 1"
        )

        # Hive connector syntax (partitioned_by)
        self.validate_identity(
            "CREATE TABLE foo (a VARCHAR, b INTEGER, c DATE) WITH (PARTITIONED_BY=ARRAY['a', 'b'])"
        )
        self.validate_identity(
            'CREATE TABLE "foo" ("a" VARCHAR, "b" INTEGER, "c" DATE) WITH (PARTITIONED_BY=ARRAY[\'a\', \'b\'])',
            identify=True,
        )

        # Iceberg connector syntax (partitioning, can contain Iceberg transform expressions)
        self.validate_identity(
            "CREATE TABLE foo (a VARCHAR, b INTEGER, c DATE) WITH (PARTITIONING=ARRAY['a', 'bucket(4, b)', 'month(c)'])",
        )
        self.validate_identity(
            'CREATE TABLE "foo" ("a" VARCHAR, "b" INTEGER, "c" DATE) WITH (PARTITIONING=ARRAY[\'a\', \'bucket(4, b)\', \'month(c)\'])',
            identify=True,
        )

    def test_analyze(self):
        self.validate_identity("ANALYZE tbl")
        self.validate_identity("ANALYZE tbl WITH (prop1=val1, prop2=val2)")

    def test_json_value(self):
        self.validate_identity(
            "JSON_VALUE(jl.extra_attributes, 'lax $.amount_source' RETURNING VARCHAR)"
        )

        json_doc = """'{"item": "shoes", "price": "49.95"}'"""
        self.validate_identity(f"""SELECT JSON_VALUE({json_doc}, 'strict $.price')""")
        self.validate_identity(
            f"""SELECT JSON_VALUE({json_doc}, 'lax $.price' RETURNING DECIMAL(4, 2))"""
        )

        for on_option in ("NULL", "ERROR", "DEFAULT 1"):
            self.validate_identity(
                f"""SELECT JSON_VALUE({json_doc}, 'lax $.price' RETURNING DECIMAL(4, 2) {on_option} ON EMPTY {on_option} ON ERROR) AS price"""
            )

    def test_array_first(self):
        self.validate_identity("SELECT ARRAY_FIRST(ARRAY['a', 'b']) FROM tbl")
        self.validate_identity("SELECT ARRAY_FIRST(ARRAY['a', 'b'], x -> x = 'b') FROM tbl")

    def test_inline_udf(self):
        self.validate_identity(
            "WITH FUNCTION f(num INTEGER) RETURNS INTEGER RETURN num SELECT F(1)"
        )
        self.validate_identity(
            "WITH FUNCTION hello(name VARCHAR) RETURNS VARCHAR RETURN FORMAT('Hello %s!', name), "
            "FUNCTION bye() RETURNS VARCHAR RETURN 'Bye!' "
            "SELECT HELLO('Finn') || BYE()"
        )
        self.validate_identity(
            "WITH FUNCTION doubled(x INTEGER) RETURNS INTEGER RETURN x * 2 "
            "WITH t AS (SELECT 3 AS v) SELECT DOUBLED(v) FROM t"
        )
        self.validate_identity(
            "WITH FUNCTION f(x INTEGER) RETURNS INTEGER RETURN x "
            "WITH RECURSIVE t(n) AS (SELECT 1 AS n) SELECT F(n) FROM t"
        )
        self.validate_identity(
            """WITH FUNCTION f(num int)
    RETURNS int
    RETURN num
SELECT f(1)""",
            "WITH FUNCTION f(num INTEGER) RETURNS INTEGER RETURN num SELECT F(1)",
        )
        self.validate_identity("WITH function AS (SELECT 1 AS x) SELECT x FROM function")
        self.validate_identity("WITH function(x) AS (SELECT 1) SELECT x FROM function")

        self.validate_identity("WITH FUNCTION f() RETURNS INTEGER LANGUAGE SQL RETURN 1 SELECT F()")
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER DETERMINISTIC RETURN 1 SELECT F()"
        )
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER NOT DETERMINISTIC RETURN 1 SELECT F()"
        )
        self.assertIsInstance(
            self.validate_identity("SELECT NOT deterministic FROM t").selects[0], exp.Not
        )
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER CALLED ON NULL INPUT RETURN 1 SELECT F()"
        )
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER RETURNS NULL ON NULL INPUT RETURN 1 SELECT F()"
        )
        self.validate_identity("WITH FUNCTION f() RETURNS INTEGER COMMENT 'hi' RETURN 1 SELECT F()")

        # SECURITY and WITH (...) are part of Trino's documented function-specification
        # grammar, but real Trino rejects both for LANGUAGE SQL inline functions
        # specifically ("Security mode not supported for inline functions", "Function
        # language 'SQL' does not support properties"). These assert round-trip
        # correctness for that shared grammar, not that this exact combination executes
        # on an inline SQL UDF.
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER SECURITY DEFINER RETURN 1 SELECT F()"
        )
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER SECURITY INVOKER RETURN 1 SELECT F()"
        )
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER WITH (weight=42) RETURN 1 SELECT F()"
        )
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER LANGUAGE SQL WITH (weight=42, cost='low') "
            "RETURN 1 SELECT F()"
        )
        self.validate_identity(
            "WITH FUNCTION custom_sqrt(a INTEGER) RETURNS DOUBLE COMMENT 'Custom sqrt function' "
            "RETURNS NULL ON NULL INPUT NOT DETERMINISTIC LANGUAGE SQL SECURITY DEFINER "
            "WITH (weight=42, cost='low') RETURN a SELECT CUSTOM_SQRT(4)"
        )

    def test_inline_udf_begin_end(self):
        # https://trino.io/docs/current/udf/sql/begin.html
        self.validate_identity(
            "WITH FUNCTION meaning_of_life() RETURNS INTEGER "
            "BEGIN DECLARE a INTEGER DEFAULT 6; DECLARE b INTEGER DEFAULT 7; RETURN a * b; END "
            "SELECT MEANING_OF_LIFE()"
        )

        # https://trino.io/docs/current/udf/sql/set.html
        self.validate_identity(
            "WITH FUNCTION one() RETURNS INTEGER "
            "BEGIN DECLARE counter INTEGER DEFAULT 1; SET counter = 0; "
            "SET counter = counter + 2; SET counter = counter / counter; RETURN counter; END "
            "SELECT ONE()"
        )

        # https://trino.io/docs/current/udf/sql/declare.html - multiple identifiers can
        # share one DECLARE and type
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER "
            "BEGIN DECLARE first_name, last_name, middle_name VARCHAR(25); RETURN 1; END "
            "SELECT F()"
        )

        # BEGIN can nest; confirmed against a real Trino instance (returns 2)
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER "
            "BEGIN DECLARE x INTEGER DEFAULT 1; BEGIN SET x = x + 1; END; RETURN x; END "
            "SELECT F()"
        )

        # DECLARE isn't reserved in Trino, so it must still work as a plain identifier
        self.validate_identity("SELECT declare FROM (VALUES (1), (2)) AS t(declare)")
        self.validate_identity("WITH FUNCTION declare() RETURNS INTEGER RETURN 1 SELECT DECLARE()")

        # A Block built without begin=True (i.e. not by _parse_routine_block) must
        # not get a synthesized BEGIN
        self.assertEqual(
            exp.Block(
                expressions=[parse_one("SELECT 1"), parse_one("SELECT 2"), exp.EndStatement()]
            ).sql(dialect="trino"),
            "SELECT 1; SELECT 2; END",
        )

    def test_inline_udf_if(self):
        # https://trino.io/docs/current/udf/sql/if.html - verbatim from the docs, but
        # real Trino rejects this exact body with "Function must end in a RETURN
        # statement": its function-body check requires a literal trailing RETURN and
        # doesn't credit an IF/ELSEIF/ELSE that already returns on every branch. This
        # asserts round-trip grammar only, confirmed against a real Trino instance.
        self.validate_identity(
            "WITH FUNCTION simple_if(a BIGINT) RETURNS VARCHAR "
            "BEGIN IF a = 0 THEN RETURN 'zero'; ELSEIF a = 1 THEN RETURN 'one'; "
            "ELSE RETURN 'more than one or negative'; END IF; END "
            "SELECT SIMPLE_IF(3)"
        )

        # IF with no ELSE/ELSEIF at all; confirmed against a real Trino instance
        self.validate_identity(
            "WITH FUNCTION f(a INTEGER) RETURNS VARCHAR "
            "BEGIN IF a = 0 THEN RETURN 'zero'; END IF; RETURN 'other'; END "
            "SELECT F(1)"
        )

        # An ELSEIF chain with no final ELSE; confirmed against a real Trino instance
        self.validate_identity(
            "WITH FUNCTION f(a INTEGER) RETURNS VARCHAR "
            "BEGIN IF a = 0 THEN RETURN 'zero'; ELSEIF a = 1 THEN RETURN 'one'; END IF; "
            "RETURN 'other'; END "
            "SELECT F(1)"
        )

        # IF can nest; the trailing RETURN is required by the same real-Trino
        # completeness check noted above, confirmed to actually run and return 'a zero'
        self.validate_identity(
            "WITH FUNCTION f(a INTEGER, b INTEGER) RETURNS VARCHAR "
            "BEGIN IF a = 0 THEN IF b = 0 THEN RETURN 'both zero'; ELSE RETURN 'a zero'; "
            "END IF; ELSE RETURN 'a nonzero'; END IF; RETURN 'unreachable'; END "
            "SELECT F(1, 2)"
        )

        # Combines with DECLARE/SET, and a CASE expression nested inside a RETURN
        # doesn't get confused with the surrounding IF's own ELSE/END IF
        self.validate_identity(
            "WITH FUNCTION f(a INTEGER) RETURNS INTEGER "
            "BEGIN DECLARE result INTEGER DEFAULT 0; "
            "IF a > 0 THEN RETURN CASE WHEN a > 10 THEN 1 ELSE 2 END; "
            "ELSE SET result = -1; END IF; RETURN result; END "
            "SELECT F(1)"
        )

        # Trino's own function-body analysis rejects a NOT DETERMINISTIC declaration
        # on a body it can tell is trivially deterministic (same class of issue as the
        # SECURITY/WITH (...) note above), so this asserts round-trip grammar only.
        self.validate_identity(
            "WITH FUNCTION f() RETURNS INTEGER LANGUAGE SQL NOT DETERMINISTIC "
            "BEGIN DECLARE x INTEGER DEFAULT 1; RETURN x; END "
            "SELECT F()"
        )
        self.validate_identity(
            "WITH FUNCTION doubled(x INTEGER) RETURNS INTEGER BEGIN RETURN x * 2; END "
            "WITH t AS (SELECT 3 AS v) SELECT DOUBLED(v) FROM t"
        )

    def test_inline_udf_case(self):
        # https://trino.io/docs/current/udf/sql/case.html - verbatim from the docs
        # (operand form). The docs label the operand form "Searched case" with a
        # synopsis ending in a bare END, but that contradicts this very example,
        # which uses the operand form and ends in END CASE; confirmed against a
        # real Trino instance that only END CASE is accepted for either form.
        self.validate_identity(
            "WITH FUNCTION simple_case(a BIGINT) RETURNS VARCHAR "
            "BEGIN CASE a WHEN 0 THEN RETURN 'zero'; WHEN 1 THEN RETURN 'one'; "
            "ELSE RETURN 'more than one or negative'; END CASE; RETURN NULL; END "
            "SELECT SIMPLE_CASE(0)"
        )

        # No-operand form ("Simple case" per the docs' own, inverted labeling);
        # confirmed against a real Trino instance
        self.validate_identity(
            "WITH FUNCTION searched_case(a BIGINT) RETURNS VARCHAR "
            "BEGIN CASE WHEN a = 0 THEN RETURN 'zero'; WHEN a = 1 THEN RETURN 'one'; "
            "ELSE RETURN 'other'; END CASE; RETURN NULL; END "
            "SELECT SEARCHED_CASE(0)"
        )

        # No ELSE at all, and only a single WHEN; confirmed against a real Trino
        # instance that this falls through to the following RETURN rather than
        # erroring
        self.validate_identity(
            "WITH FUNCTION no_else(a BIGINT) RETURNS VARCHAR "
            "BEGIN CASE a WHEN 0 THEN RETURN 'zero'; END CASE; RETURN 'fallthrough'; END "
            "SELECT NO_ELSE(0)"
        )

        # No-operand form, no ELSE; confirmed against a real Trino instance
        self.validate_identity(
            "WITH FUNCTION no_operand_no_else(a INTEGER) RETURNS VARCHAR "
            "BEGIN CASE WHEN a = 0 THEN RETURN 'zero'; END CASE; RETURN 'other'; END "
            "SELECT NO_OPERAND_NO_ELSE(1)"
        )

        # CASE can nest, in both the operand and no-operand forms; confirmed to
        # actually run and return 'a0b0'/'a0bN'/'aN' against a real Trino instance
        self.validate_identity(
            "WITH FUNCTION nested_case(a BIGINT, b BIGINT) RETURNS VARCHAR "
            "BEGIN DECLARE result VARCHAR; "
            "CASE WHEN a = 0 THEN CASE b WHEN 0 THEN SET result = 'a0b0'; "
            "ELSE SET result = 'a0bN'; END CASE; ELSE SET result = 'aN'; END CASE; "
            "RETURN result; END "
            "SELECT NESTED_CASE(0, 0)"
        )

        # Combines with IF, and a CASE expression nested inside a SET doesn't get
        # confused with the surrounding CASE statement's own ELSE/END CASE
        self.validate_identity(
            "WITH FUNCTION mix(a INTEGER) RETURNS INTEGER "
            "BEGIN DECLARE result INTEGER DEFAULT 0; "
            "CASE WHEN a > 0 THEN IF a > 10 THEN SET result = 1; ELSE SET result = 2; END IF; "
            "ELSE SET result = CASE WHEN a < -10 THEN -1 ELSE -2 END; END CASE; "
            "RETURN result; END "
            "SELECT MIX(1)"
        )

        # CASE as the literal last statement, with nothing after it before the
        # enclosing END; real Trino rejects this body with "Function must end in
        # a RETURN statement" - the same function-body completeness check noted
        # on the IF phase, which requires a literal trailing RETURN and doesn't
        # credit a CASE that already returns on every branch. This asserts
        # round-trip grammar only, confirmed against a real Trino instance.
        self.validate_identity(
            "WITH FUNCTION last_stmt(a INTEGER) RETURNS VARCHAR "
            "BEGIN CASE a WHEN 0 THEN RETURN 'zero'; ELSE RETURN 'other'; END CASE; END "
            "SELECT LAST_STMT(1)"
        )

    def test_inline_udf_while(self):
        # https://trino.io/docs/current/udf/sql/while.html - verbatim from the docs
        # (wrapped in a function since the docs show it as a body fragment);
        # confirmed against a real Trino instance (returns 625 = 5^4)
        # (renamed from the docs' own function name to avoid colliding with the
        # builtin two-argument POWER function)
        self.validate_identity(
            "WITH FUNCTION npow(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE r BIGINT DEFAULT 1; DECLARE p BIGINT DEFAULT n; "
            "WHILE p > 1 DO SET r = r * n; SET p = p - 1; END WHILE; "
            "RETURN r; END "
            "SELECT NPOW(5)"
        )

        # The optional `label :` prefix names the block for ITERATE/LEAVE (not
        # yet supported); confirmed to still return 625 against a real Trino
        # instance, and that a repeated label after END WHILE is rejected
        self.validate_identity(
            "WITH FUNCTION npow_labeled(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE r BIGINT DEFAULT 1; DECLARE p BIGINT DEFAULT n; "
            "abc: WHILE p > 1 DO SET r = r * n; SET p = p - 1; END WHILE; "
            "RETURN r; END "
            "SELECT NPOW_LABELED(5)"
        )

        # WHILE as the literal last statement, with nothing after it before the
        # enclosing END; real Trino rejects this body with "Function must end in
        # a RETURN statement", the same function-body completeness check noted
        # on the IF and CASE phases. This asserts round-trip grammar only,
        # confirmed against a real Trino instance.
        self.validate_identity(
            "WITH FUNCTION last_stmt(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE r BIGINT DEFAULT 1; WHILE r < n DO SET r = r + 1; END WHILE; END "
            "SELECT LAST_STMT(5)"
        )

        # WHILE can nest, and combines with IF and CASE; confirmed against a real
        # Trino instance to actually run and return 4 for count_evens(3)
        self.validate_identity(
            "WITH FUNCTION count_evens(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE total BIGINT DEFAULT 0; DECLARE i BIGINT DEFAULT 0; "
            "WHILE i < n DO SET total = total + 1; IF i = 0 THEN SET total = total + 1; END IF; "
            "SET i = i + 1; END WHILE; "
            "RETURN total; END "
            "SELECT COUNT_EVENS(3)"
        )
        self.validate_identity(
            "WITH FUNCTION while_case(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE i BIGINT DEFAULT 0; "
            "WHILE i < n DO CASE WHEN i = 0 THEN SET i = 1; ELSE SET i = i + 1; END CASE; END WHILE; "
            "RETURN i; END "
            "SELECT WHILE_CASE(3)"
        )

    def test_inline_udf_loop_repeat(self):
        # https://trino.io/docs/current/udf/sql/loop.html - verbatim from the docs
        # (wrapped as a function invocation); confirmed against a real Trino
        # instance to return 10, 20, 30 for step values of 10, 20, 30
        loop_ast = self.validate_identity(
            "WITH FUNCTION to_one_hundred(start_value BIGINT, step BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE count BIGINT DEFAULT 0; DECLARE current BIGINT DEFAULT 0; "
            "SET current = start_value; "
            "abc: LOOP IF current >= 100 THEN LEAVE abc; END IF; "
            "SET count = count + 1; SET current = current + step; END LOOP; "
            "RETURN count; END "
            "SELECT TO_ONE_HUNDRED(0, 10)"
        )

        # https://trino.io/docs/current/udf/sql/repeat.html - verbatim from the
        # docs; unlike WHILE/LOOP the body always runs at least once, confirmed
        # against a real Trino instance to return 10, 10, 11, 12, 13 for inputs
        # 5, 9, 10, 11, 12
        self.validate_identity(
            "WITH FUNCTION test_repeat(a BIGINT) RETURNS BIGINT "
            "BEGIN REPEAT SET a = a + 1; UNTIL a >= 10 END REPEAT; RETURN a; END "
            "SELECT TEST_REPEAT(5)"
        )

        # https://trino.io/docs/current/udf/sql/iterate.html - verbatim from the
        # docs (renamed from the docs' own function name to avoid colliding with
        # the builtin COUNT aggregate); confirmed against a real Trino instance
        # to return 7
        repeat_ast = self.validate_identity(
            "WITH FUNCTION iter_count() RETURNS BIGINT "
            "BEGIN DECLARE a BIGINT DEFAULT 0; DECLARE b BIGINT DEFAULT 0; "
            "top: REPEAT SET a = a + 1; IF a <= 3 THEN ITERATE top; END IF; SET b = b + 1; "
            "UNTIL a >= 10 END REPEAT; "
            "RETURN b; END "
            "SELECT ITER_COUNT()"
        )

        # LOOP can nest inside WHILE, and LEAVE targets the innermost matching
        # label; confirmed against a real Trino instance to return 6 (2 outer
        # iterations x 3 inner increments each)
        self.validate_identity(
            "WITH FUNCTION nested_test(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE total BIGINT DEFAULT 0; DECLARE i BIGINT DEFAULT 0; DECLARE j BIGINT DEFAULT 0; "
            "outer_loop: WHILE i < n DO SET j = 0; "
            "inner_loop: LOOP IF j >= 3 THEN LEAVE inner_loop; END IF; "
            "SET total = total + 1; SET j = j + 1; END LOOP; "
            "SET i = i + 1; END WHILE; "
            "RETURN total; END "
            "SELECT NESTED_TEST(2)"
        )

        # LOOP/REPEAT as the literal last statement, with nothing after before
        # the enclosing END; real Trino rejects this body with "Function must
        # end in a RETURN statement", the same function-body completeness check
        # noted on the IF/CASE/WHILE phases. This asserts round-trip grammar
        # only, confirmed against a real Trino instance.
        self.validate_identity(
            "WITH FUNCTION last_stmt(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE i BIGINT DEFAULT 0; "
            "top: LOOP IF i >= n THEN LEAVE top; END IF; SET i = i + 1; END LOOP; END "
            "SELECT LAST_STMT(5)"
        )

        # `ITERATE`/`LEAVE` are themselves valid label names, confirmed against
        # a real Trino instance to still return 5 for both; the label lookahead
        # has to be checked before ITERATE/LEAVE are matched as keywords, or
        # this misparses as a bare (invalid) ITERATE/LEAVE statement
        self.validate_identity(
            "WITH FUNCTION label_test(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE i BIGINT DEFAULT 0; "
            "iterate: LOOP IF i >= n THEN LEAVE iterate; END IF; SET i = i + 1; END LOOP; "
            "RETURN i; END "
            "SELECT LABEL_TEST(5)"
        )
        self.validate_identity(
            "WITH FUNCTION label_test2(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE i BIGINT DEFAULT 0; "
            "leave: LOOP IF i >= n THEN LEAVE leave; END IF; SET i = i + 1; END LOOP; "
            "RETURN i; END "
            "SELECT LABEL_TEST2(5)"
        )

        # Any non-reserved keyword is a valid label name, so the label lookahead
        # has to run before the statement keywords (e.g. SET) are matched;
        # confirmed against a real Trino instance to return 5
        self.validate_identity(
            "WITH FUNCTION label_test3(n BIGINT) RETURNS BIGINT "
            "BEGIN DECLARE i BIGINT DEFAULT 0; "
            "set: LOOP IF i >= n THEN LEAVE set; END IF; SET i = i + 1; END LOOP; "
            "RETURN i; END "
            "SELECT LABEL_TEST3(5)"
        )

        # The base generator warns on these Trino-specific routine statements
        # instead of crashing
        for node in (
            *loop_ast.find_all(exp.LoopBlock, exp.Leave),
            *repeat_ast.find_all(exp.RepeatBlock, exp.Iterate),
        ):
            with self.assertLogs(generator_logger, level="WARNING") as cm:
                self.assertEqual(node.sql(), "")

            self.assertIn("Unsupported", cm.output[0])



from __future__ import annotations


from sqlglot.dialects.trino import Trino
from sqlglot.generators.dune import DuneGenerator
from sqlglot.parsers.dune import DuneParser


class Dune(Trino):
    Parser = DuneParser

    class Tokenizer(Trino.Tokenizer):
        HEX_STRINGS = ["0x", ("X'", "'")]

    Generator = DuneGenerator



from __future__ import annotations

import typing as t

from sqlglot import exp, generator, transforms
from sqlglot.dialects.dialect import (
    arrow_json_extract_sql,
    build_date_delta,
    build_date_delta_with_interval,
    date_add_interval_sql,
    datestrtodate_sql,
    length_or_char_length_sql,
    max_or_greatest,
    min_or_least,
    no_ilike_sql,
    no_paren_current_date_sql,
    no_pivot_sql,
    no_tablesample_sql,
    no_trycast_sql,
    remove_ts_or_ds_to_date,
    rename_func,
    strposition_sql,
    unit_to_var,
    trim_sql,
    timestrtotime_sql,
)
from sqlglot.generator import unsupported_args
from collections import defaultdict


def _date_trunc_sql(self: MySQLGenerator, expression: exp.DateTrunc) -> str:
    expr = self.sql(expression, "this")
    unit_expr = expression.args.get("unit")
    unit = (
        self.weekstart_name(unit_expr)
        if isinstance(unit_expr, exp.WeekStart)
        else expression.text("unit").upper()
    )

    if unit == "WEEK":
        concat = f"CONCAT(YEAR({expr}), ' ', WEEK({expr}, 1), ' 1')"
        date_format = "%Y %u %w"
    elif unit == "MONTH":
        concat = f"CONCAT(YEAR({expr}), ' ', MONTH({expr}), ' 1')"
        date_format = "%Y %c %e"
    elif unit == "QUARTER":
        concat = f"CONCAT(YEAR({expr}), ' ', QUARTER({expr}) * 3 - 2, ' 1')"
        date_format = "%Y %c %e"
    elif unit == "YEAR":
        concat = f"CONCAT(YEAR({expr}), ' 1 1')"
        date_format = "%Y %c %e"
    else:
        if unit != "DAY":
            self.unsupported(f"Unexpected interval unit: {unit}")
        return self.func("DATE", expr)

    return self.func("STR_TO_DATE", concat, f"'{date_format}'")


def _str_to_date_sql(
    self: MySQLGenerator, expression: exp.StrToDate | exp.StrToTime | exp.TsOrDsToDate
) -> str:
    return self.func("STR_TO_DATE", expression.this, self.format_time(expression))


def _unix_to_time_sql(self: MySQLGenerator, expression: exp.UnixToTime) -> str:
    scale = expression.args.get("scale")
    timestamp = expression.this

    if scale in (None, exp.UnixToTime.SECONDS):
        return self.func("FROM_UNIXTIME", timestamp, self.format_time(expression))

    return self.func(
        "FROM_UNIXTIME",
        exp.Div(this=timestamp, expression=exp.func("POW", 10, scale)),
        self.format_time(expression),
    )


def date_add_sql(
    kind: str,
) -> t.Callable[[generator.Generator, exp.Expr], str]:
    def func(self: generator.Generator, expression: exp.Expr) -> str:
        return self.func(
            f"DATE_{kind}",
            expression.this,
            exp.Interval(this=expression.expression, unit=unit_to_var(expression)),
        )

    return func


_MAKE_INTERVAL_UNIT_ALIASES = {
    "years": "year",
    "months": "month",
    "weeks": "week",
    "days": "day",
    "hours": "hour",
    "minutes": "minute",
    "mins": "minute",
    "seconds": "second",
    "secs": "second",
}


def _ts_or_ds_to_date_sql(self: MySQLGenerator, expression: exp.TsOrDsToDate) -> str:
    time_format = expression.args.get("format")
    return _str_to_date_sql(self, expression) if time_format else self.func("DATE", expression.this)


class MySQLGenerator(generator.Generator):
    SELECT_KINDS: tuple[str, ...] = ()
    TRY_SUPPORTED = False
    SUPPORTS_UESCAPE = False
    SUPPORTS_DECODE_CASE = False
    SUPPORTS_MODIFY_COLUMN = True
    SUPPORTS_CHANGE_COLUMN = True
    SUPPORTS_ALTER_COLUMN_NULLABILITY = True

    AFTER_HAVING_MODIFIER_TRANSFORMS = generator.AFTER_HAVING_MODIFIER_TRANSFORMS

    INTERVAL_ALLOWS_PLURAL_FORM = False
    LOCKING_READS_SUPPORTED = True
    NULL_ORDERING_SUPPORTED: bool | None = None
    JOIN_HINTS = False
    TABLE_HINTS = True
    DUPLICATE_KEY_UPDATE_WITH_SET = False
    QUERY_HINT_SEP = " "
    VALUES_AS_TABLE = False
    NVL2_SUPPORTED = False
    LAST_DAY_SUPPORTS_DATE_PART = False
    JSON_TYPE_REQUIRED_FOR_EXTRACTION = True
    JSON_PATH_BRACKETED_KEY_SUPPORTED = False
    JSON_KEY_VALUE_PAIR_SEP = ","
    SUPPORTS_TO_NUMBER = False
    PARSE_JSON_NAME: str | None = None
    PAD_FILL_PATTERN_IS_REQUIRED = True
    WRAP_DERIVED_VALUES = False
    VARCHAR_REQUIRES_SIZE = True
    SUPPORTS_MEDIAN = False
    UPDATE_STATEMENT_SUPPORTS_FROM = False

    TRANSFORMS = {
        **generator.Generator.TRANSFORMS,
        exp.ArrayAgg: rename_func("GROUP_CONCAT"),
        exp.BitwiseAndAgg: rename_func("BIT_AND"),
        exp.BitwiseOrAgg: rename_func("BIT_OR"),
        exp.BitwiseXorAgg: rename_func("BIT_XOR"),
        exp.BitwiseCount: rename_func("BIT_COUNT"),
        exp.Chr: lambda self, e: self.chr_sql(e, "CHAR"),
        exp.CurrentDate: no_paren_current_date_sql,
        exp.CurrentVersion: rename_func("VERSION"),
        exp.DateDiff: remove_ts_or_ds_to_date(
            lambda self, e: self.func("DATEDIFF", e.this, e.expression), ("this", "expression")
        ),
        exp.DateAdd: remove_ts_or_ds_to_date(date_add_sql("ADD")),
        exp.DateStrToDate: datestrtodate_sql,
        exp.DateSub: remove_ts_or_ds_to_date(date_add_sql("SUB")),
        exp.DateTrunc: _date_trunc_sql,
        exp.Day: remove_ts_or_ds_to_date(),
        exp.DayOfMonth: remove_ts_or_ds_to_date(rename_func("DAYOFMONTH")),
        exp.DayOfWeek: remove_ts_or_ds_to_date(rename_func("DAYOFWEEK")),
        exp.DayOfYear: remove_ts_or_ds_to_date(rename_func("DAYOFYEAR")),
        exp.GroupConcat: lambda self, e: (
            f"""GROUP_CONCAT({self.sql(e, "this")} SEPARATOR {self.sql(e, "separator") or "','"})"""
        ),
        exp.ILike: no_ilike_sql,
        exp.JSONExtractScalar: arrow_json_extract_sql,
        exp.Length: length_or_char_length_sql,
        exp.LogicalOr: rename_func("MAX"),
        exp.LogicalAnd: rename_func("MIN"),
        exp.Max: max_or_greatest,
        exp.Min: min_or_least,
        exp.Month: remove_ts_or_ds_to_date(),
        exp.NullSafeEQ: lambda self, e: self.binary(e, "<=>"),
        exp.NullSafeNEQ: lambda self, e: f"NOT {self.binary(e, '<=>')}",
        exp.NumberToStr: rename_func("FORMAT"),
        exp.Pivot: no_pivot_sql,
        exp.Select: transforms.preprocess(
            [
                transforms.eliminate_distinct_on,
                transforms.eliminate_semi_and_anti_joins,
                transforms.eliminate_qualify,
                transforms.eliminate_full_outer_join,
                transforms.unnest_generate_date_array_using_recursive_cte,
            ]
        ),
        exp.StrPosition: lambda self, e: strposition_sql(
            self, e, func_name="LOCATE", supports_position=True
        ),
        exp.StrToDate: _str_to_date_sql,
        exp.StrToTime: _str_to_date_sql,
        exp.Stuff: rename_func("INSERT"),
        exp.SessionUser: lambda *_: "SESSION_USER()",
        exp.TableSample: no_tablesample_sql,
        exp.TimeFromParts: rename_func("MAKETIME"),
        exp.TimestampAdd: date_add_interval_sql("DATE", "ADD"),
        exp.TimestampDiff: lambda self, e: self.func(
            "TIMESTAMPDIFF", unit_to_var(e), e.expression, e.this
        ),
        exp.TimestampSub: date_add_interval_sql("DATE", "SUB"),
        exp.TimeStrToUnix: rename_func("UNIX_TIMESTAMP"),
        exp.TimeStrToTime: lambda self, e: timestrtotime_sql(
            self,
            e,
            include_precision=not e.args.get("zone"),
        ),
        exp.TimeToStr: remove_ts_or_ds_to_date(
            lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e))
        ),
        exp.Trim: trim_sql,
        exp.Trunc: rename_func("TRUNCATE"),
        exp.TryCast: no_trycast_sql,
        exp.TsOrDsAdd: date_add_sql("ADD"),
        exp.TsOrDsDiff: lambda self, e: self.func("DATEDIFF", e.this, e.expression),
        exp.TsOrDsToDate: _ts_or_ds_to_date_sql,
        exp.Unicode: lambda self, e: f"ORD(CONVERT({self.sql(e.this)} USING utf32))",
        exp.UnixToTime: _unix_to_time_sql,
        exp.Week: remove_ts_or_ds_to_date(),
        exp.WeekOfYear: remove_ts_or_ds_to_date(rename_func("WEEKOFYEAR")),
        exp.Year: remove_ts_or_ds_to_date(),
        exp.UtcTimestamp: rename_func("UTC_TIMESTAMP"),
        exp.UtcTime: rename_func("UTC_TIME"),
    }

    UNSIGNED_TYPE_MAPPING: t.ClassVar = {
        exp.DType.UBIGINT: "BIGINT",
        exp.DType.UINT: "INT",
        exp.DType.UMEDIUMINT: "MEDIUMINT",
        exp.DType.USMALLINT: "SMALLINT",
        exp.DType.UTINYINT: "TINYINT",
        exp.DType.UDECIMAL: "DECIMAL",
        exp.DType.UDOUBLE: "DOUBLE",
    }

    TIMESTAMP_TYPE_MAPPING: t.ClassVar = {
        exp.DType.DATETIME2: "DATETIME",
        exp.DType.SMALLDATETIME: "DATETIME",
        exp.DType.TIMESTAMP: "DATETIME",
        exp.DType.TIMESTAMPNTZ: "DATETIME",
        exp.DType.TIMESTAMPTZ: "TIMESTAMP",
        exp.DType.TIMESTAMPLTZ: "TIMESTAMP",
    }

    TYPE_MAPPING: t.ClassVar = {
        exp.DType.NCHAR: "CHAR",
        exp.DType.NVARCHAR: "VARCHAR",
        exp.DType.INET: "INET",
        exp.DType.ROWVERSION: "VARBINARY",
        exp.DType.UBIGINT: "BIGINT",
        exp.DType.UINT: "INT",
        exp.DType.UMEDIUMINT: "MEDIUMINT",
        exp.DType.USMALLINT: "SMALLINT",
        exp.DType.UTINYINT: "TINYINT",
        exp.DType.UDECIMAL: "DECIMAL",
        exp.DType.UDOUBLE: "DOUBLE",
        exp.DType.DATETIME2: "DATETIME",
        exp.DType.SMALLDATETIME: "DATETIME",
        exp.DType.TIMESTAMP: "DATETIME",
        exp.DType.TIMESTAMPNTZ: "DATETIME",
        exp.DType.TIMESTAMPTZ: "TIMESTAMP",
        exp.DType.TIMESTAMPLTZ: "TIMESTAMP",
    }

    PROPERTIES_LOCATION: t.ClassVar = {
        **generator.Generator.PROPERTIES_LOCATION,
        exp.TransientProperty: exp.Properties.Location.UNSUPPORTED,
        exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
        exp.PartitionedByProperty: exp.Properties.Location.UNSUPPORTED,
        exp.PartitionByRangeProperty: exp.Properties.Location.POST_SCHEMA,
        exp.PartitionByListProperty: exp.Properties.Location.POST_SCHEMA,
    }

    LIMIT_FETCH = "LIMIT"

    LIMIT_ONLY_LITERALS = True

    CHAR_CAST_MAPPING: t.ClassVar = dict.fromkeys(
        (
            exp.DType.LONGTEXT,
            exp.DType.LONGBLOB,
            exp.DType.MEDIUMBLOB,
            exp.DType.MEDIUMTEXT,
            exp.DType.TEXT,
            exp.DType.TINYBLOB,
            exp.DType.TINYTEXT,
            exp.DType.VARCHAR,
        ),
        "CHAR",
    )
    SIGNED_CAST_MAPPING: t.ClassVar = dict.fromkeys(
        (
            exp.DType.BIGINT,
            exp.DType.BOOLEAN,
            exp.DType.INT,
            exp.DType.SMALLINT,
            exp.DType.TINYINT,
            exp.DType.MEDIUMINT,
        ),
        "SIGNED",
    )

    # MySQL doesn't support many datatypes in cast.
    # https://dev.mysql.com/doc/refman/8.0/en/cast-functions.html#function_cast
    CAST_MAPPING: t.ClassVar = {
        exp.DType.LONGTEXT: "CHAR",
        exp.DType.LONGBLOB: "CHAR",
        exp.DType.MEDIUMBLOB: "CHAR",
        exp.DType.MEDIUMTEXT: "CHAR",
        exp.DType.TEXT: "CHAR",
        exp.DType.TINYBLOB: "CHAR",
        exp.DType.TINYTEXT: "CHAR",
        exp.DType.VARCHAR: "CHAR",
        exp.DType.BIGINT: "SIGNED",
        exp.DType.BOOLEAN: "SIGNED",
        exp.DType.INT: "SIGNED",
        exp.DType.SMALLINT: "SIGNED",
        exp.DType.TINYINT: "SIGNED",
        exp.DType.MEDIUMINT: "SIGNED",
        exp.DType.UBIGINT: "UNSIGNED",
    }

    TIMESTAMP_FUNC_TYPES: t.ClassVar = {
        exp.DType.TIMESTAMPTZ,
        exp.DType.TIMESTAMPLTZ,
    }

    # https://dev.mysql.com/doc/refman/8.0/en/keywords.html
    RESERVED_KEYWORDS = {
        "accessible",
        "add",
        "all",
        "alter",
        "analyze",
        "and",
        "as",
        "asc",
        "asensitive",
        "before",
        "between",
        "bigint",
        "binary",
        "blob",
        "both",
        "by",
        "call",
        "cascade",
        "case",
        "change",
        "char",
        "character",
        "check",
        "collate",
        "column",
        "condition",
        "constraint",
        "continue",
        "convert",
        "create",
        "cross",
        "cube",
        "cume_dist",
        "current_date",
        "current_time",
        "current_timestamp",
        "current_user",
        "cursor",
        "database",
        "databases",
        "day_hour",
        "day_microsecond",
        "day_minute",
        "day_second",
        "dec",
        "decimal",
        "declare",
        "default",
        "delayed",
        "delete",
        "dense_rank",
        "desc",
        "describe",
        "deterministic",
        "distinct",
        "distinctrow",
        "div",
        "double",
        "drop",
        "dual",
        "each",
        "else",
        "elseif",
        "empty",
        "enclosed",
        "escaped",
        "except",
        "exists",
        "exit",
        "explain",
        "false",
        "fetch",
        "first_value",
        "float",
        "float4",
        "float8",
        "for",
        "force",
        "foreign",
        "from",
        "fulltext",
        "function",
        "generated",
        "get",
        "grant",
        "group",
        "grouping",
        "groups",
        "having",
        "high_priority",
        "hour_microsecond",
        "hour_minute",
        "hour_second",
        "if",
        "ignore",
        "in",
        "index",
        "infile",
        "inner",
        "inout",
        "insensitive",
        "insert",
        "int",
        "int1",
        "int2",
        "int3",
        "int4",
        "int8",
        "integer",
        "intersect",
        "interval",
        "into",
        "io_after_gtids",
        "io_before_gtids",
        "is",
        "iterate",
        "join",
        "json_table",
        "key",
        "keys",
        "kill",
        "lag",
        "last_value",
        "lateral",
        "lead",
        "leading",
        "leave",
        "left",
        "like",
        "limit",
        "linear",
        "lines",
        "load",
        "localtime",
        "localtimestamp",
        "lock",
        "long",
        "longblob",
        "longtext",
        "loop",
        "low_priority",
        "master_bind",
        "master_ssl_verify_server_cert",
        "match",
        "maxvalue",
        "mediumblob",
        "mediumint",
        "mediumtext",
        "middleint",
        "minute_microsecond",
        "minute_second",
        "mod",
        "modifies",
        "natural",
        "not",
        "no_write_to_binlog",
        "nth_value",
        "ntile",
        "null",
        "numeric",
        "of",
        "on",
        "optimize",
        "optimizer_costs",
        "option",
        "optionally",
        "or",
        "order",
        "out",
        "outer",
        "outfile",
        "over",
        "partition",
        "percent_rank",
        "precision",
        "primary",
        "procedure",
        "purge",
        "range",
        "rank",
        "read",
        "reads",
        "read_write",
        "real",
        "recursive",
        "references",
        "regexp",
        "release",
        "rename",
        "repeat",
        "replace",
        "require",
        "resignal",
        "restrict",
        "return",
        "revoke",
        "right",
        "rlike",
        "row",
        "rows",
        "row_number",
        "schema",
        "schemas",
        "second_microsecond",
        "select",
        "sensitive",
        "separator",
        "set",
        "show",
        "signal",
        "smallint",
        "spatial",
        "specific",
        "sql",
        "sqlexception",
        "sqlstate",
        "sqlwarning",
        "sql_big_result",
        "sql_calc_found_rows",
        "sql_small_result",
        "ssl",
        "starting",
        "stored",
        "straight_join",
        "system",
        "table",
        "terminated",
        "then",
        "tinyblob",
        "tinyint",
        "tinytext",
        "to",
        "trailing",
        "trigger",
        "true",
        "undo",
        "union",
        "unique",
        "unlock",
        "unsigned",
        "update",
        "usage",
        "use",
        "using",
        "utc_date",
        "utc_time",
        "utc_timestamp",
        "values",
        "varbinary",
        "varchar",
        "varcharacter",
        "varying",
        "virtual",
        "when",
        "where",
        "while",
        "window",
        "with",
        "write",
        "xor",
        "year_month",
        "zerofill",
    }

    SQL_SECURITY_VIEW_LOCATION: t.ClassVar = exp.Properties.Location.POST_CREATE

    def makeinterval_sql(self: MySQLGenerator, expression: exp.MakeInterval) -> str:
        intervals: list[exp.Interval] = []
        for arg_key, value in expression.args.items():
            if value is None:
                continue

            if isinstance(value, exp.Kwarg):
                unit_name = _MAKE_INTERVAL_UNIT_ALIASES.get(
                    value.this.name.lower(), value.this.name.lower()
                )
                value = value.expression
            else:
                unit_name = arg_key

            intervals.append(exp.Interval(this=value.copy(), unit=exp.var(unit_name.upper())))

        if not intervals:
            return self.function_fallback_sql(expression)

        parent = expression.parent
        sep = " - " if isinstance(parent, exp.Sub) and parent.expression is expression else " + "

        return sep.join(self.sql(interval) for interval in intervals)

    def locate_properties(self, properties: exp.Properties) -> defaultdict:
        locations = super().locate_properties(properties)

        # MySQL puts SQL SECURITY before VIEW but after the schema for functions/procedures
        if isinstance(create := properties.parent, exp.Create) and create.kind == "VIEW":
            post_schema = locations[exp.Properties.Location.POST_SCHEMA]
            for i, p in enumerate(post_schema):
                if isinstance(p, exp.SqlSecurityProperty):
                    post_schema.pop(i)
                    locations[self.SQL_SECURITY_VIEW_LOCATION].append(p)
                    break

        return locations

    def computedcolumnconstraint_sql(self, expression: exp.ComputedColumnConstraint) -> str:
        persisted = "STORED" if expression.args.get("persisted") else "VIRTUAL"
        return f"GENERATED ALWAYS AS ({self.sql(expression.this.unnest())}) {persisted}"

    def array_sql(self, expression: exp.Array) -> str:
        self.unsupported("Arrays are not supported by MySQL")
        return self.function_fallback_sql(expression)

    def arraycontainsall_sql(self, expression: exp.ArrayContainsAll) -> str:
        self.unsupported("Array operations are not supported by MySQL")
        return self.function_fallback_sql(expression)

    def arraycontainedby_sql(self, expression: exp.ArrayContainedBy) -> str:
        self.unsupported("Array operations are not supported by MySQL")
        return self.function_fallback_sql(expression)

    def dpipe_sql(self, expression: exp.DPipe) -> str:
        return self.func("CONCAT", *expression.flatten())

    def extract_sql(self, expression: exp.Extract) -> str:
        unit = expression.name
        if unit and unit.lower() == "epoch":
            return self.func("UNIX_TIMESTAMP", expression.expression)

        return super().extract_sql(expression)

    def datatype_sql(self, expression: exp.DataType) -> str:
        if (
            self.VARCHAR_REQUIRES_SIZE
            and expression.is_type(exp.DType.VARCHAR)
            and not expression.expressions
        ):
            # `VARCHAR` must always have a size - if it doesn't, we always generate `TEXT`
            return "TEXT"

        # https://dev.mysql.com/doc/refman/8.0/en/numeric-type-syntax.html
        result = super().datatype_sql(expression)
        if expression.this in self.UNSIGNED_TYPE_MAPPING:
            result = f"{result} UNSIGNED"

        return result

    def jsonarraycontains_sql(self, expression: exp.JSONArrayContains) -> str:
        return f"{self.sql(expression, 'this')} MEMBER OF({self.sql(expression, 'expression')})"

    def cast_sql(self, expression: exp.Cast, safe_prefix: str | None = None) -> str:
        if expression.to.this in self.TIMESTAMP_FUNC_TYPES:
            return self.func("TIMESTAMP", expression.this)

        to = self.CAST_MAPPING.get(expression.to.this)

        if to:
            expression.to.set("this", to)
        return super().cast_sql(expression)

    def show_sql(self, expression: exp.Show) -> str:
        this = f" {expression.name}"
        full = " FULL" if expression.args.get("full") else ""
        global_ = " GLOBAL" if expression.args.get("global_") else ""

        target = self.sql(expression, "target")
        target = f" {target}" if target else ""
        if expression.name in ("COLUMNS", "INDEX"):
            target = f" FROM{target}"
        elif expression.name == "GRANTS":
            target = f" FOR{target}"
        elif expression.name in ("LINKS", "PARTITIONS"):
            target = f" ON{target}" if target else ""
        elif expression.name == "PROJECTIONS":
            target = f" ON TABLE{target}" if target else ""

        db = self._prefixed_sql("FROM", expression, "db")

        like = self._prefixed_sql("LIKE", expression, "like")
        where = self.sql(expression, "where")

        types = self.expressions(expression, key="types")
        types = f" {types}" if types else types
        query = self._prefixed_sql("FOR QUERY", expression, "query")

        if expression.name == "PROFILE":
            offset = self._prefixed_sql("OFFSET", expression, "offset")
            limit = self._prefixed_sql("LIMIT", expression, "limit")
        else:
            offset = ""
            limit = self._oldstyle_limit_sql(expression)

        log = self._prefixed_sql("IN", expression, "log")
        position = self._prefixed_sql("FROM", expression, "position")

        channel = self._prefixed_sql("FOR CHANNEL", expression, "channel")

        if expression.name == "ENGINE":
            mutex_or_status = " MUTEX" if expression.args.get("mutex") else " STATUS"
        else:
            mutex_or_status = ""

        for_table = self._prefixed_sql("FOR TABLE", expression, "for_table")
        for_group = self._prefixed_sql("FOR GROUP", expression, "for_group")
        for_user = self._prefixed_sql("FOR USER", expression, "for_user")
        for_role = self._prefixed_sql("FOR ROLE", expression, "for_role")
        into_outfile = self._prefixed_sql("INTO OUTFILE", expression, "into_outfile")
        json = " JSON" if expression.args.get("json") else ""

        return f"SHOW{full}{global_}{this}{json}{target}{for_table}{types}{db}{query}{log}{position}{channel}{mutex_or_status}{like}{where}{offset}{limit}{for_group}{for_user}{for_role}{into_outfile}"

    def alterrename_sql(self, expression: exp.AlterRename, include_to: bool = True) -> str:
        """To avoid TO keyword in ALTER ... RENAME statements.
        It's moved from Doris, because it's the same for all MySQL, Doris, and StarRocks.
        """
        return super().alterrename_sql(expression, include_to=False)

    def altercolumn_sql(self, expression: exp.AlterColumn) -> str:
        dtype = self.sql(expression, "dtype")
        if not dtype:
            return super().altercolumn_sql(expression)

        if expression.args.get("exists"):
            self.unsupported("ALTER COLUMN IF EXISTS is not supported by this dialect")

        this = self.sql(expression, "this")
        null_constraint = self._alter_column_null_constraint_sql(expression)
        return f"MODIFY COLUMN {this} {dtype}{null_constraint}"

    def _prefixed_sql(self, prefix: str, expression: exp.Expr, arg: str) -> str:
        sql = self.sql(expression, arg)
        return f" {prefix} {sql}" if sql else ""

    def _oldstyle_limit_sql(self, expression: exp.Show) -> str:
        limit = self.sql(expression, "limit")
        offset = self.sql(expression, "offset")
        if limit:
            limit_offset = f"{offset}, {limit}" if offset else limit
            return f" LIMIT {limit_offset}"
        return ""

    def timestamptrunc_sql(self, expression: exp.TimestampTrunc) -> str:
        unit = expression.args.get("unit")
        if isinstance(unit, exp.WeekStart):
            unit = exp.var(self.weekstart_name(unit))

        # Pick an old-enough date to avoid negative timestamp diffs
        start_ts = "'0000-01-01 00:00:00'"

        # Source: https://stackoverflow.com/a/32955740
        timestamp_diff = build_date_delta(exp.TimestampDiff)([unit, start_ts, expression.this])
        interval = exp.Interval(this=timestamp_diff, unit=unit)
        dateadd = build_date_delta_with_interval(exp.DateAdd)([start_ts, interval])

        return self.sql(dateadd)

    def converttimezone_sql(self, expression: exp.ConvertTimezone) -> str:
        from_tz = expression.args.get("source_tz")
        to_tz = expression.args.get("target_tz")
        dt = expression.args.get("timestamp")

        return self.func("CONVERT_TZ", dt, from_tz, to_tz)

    def attimezone_sql(self, expression: exp.AtTimeZone) -> str:
        self.unsupported("AT TIME ZONE is not supported by MySQL")
        return self.sql(expression.this)

    def isascii_sql(self, expression: exp.IsAscii) -> str:
        return f"REGEXP_LIKE({self.sql(expression.this)}, '^[[:ascii:]]*$')"

    def ignorenulls_sql(self, expression: exp.IgnoreNulls) -> str:
        # https://dev.mysql.com/doc/refman/8.4/en/window-function-descriptions.html
        self.unsupported("MySQL does not support IGNORE NULLS.")
        return self.sql(expression.this)

    @unsupported_args("this")
    def currentschema_sql(self, expression: exp.CurrentSchema) -> str:
        return self.func("SCHEMA")

    def partition_sql(self, expression: exp.Partition) -> str:
        parent = expression.parent
        if isinstance(parent, (exp.PartitionByRangeProperty, exp.PartitionByListProperty)):
            return self.expressions(expression, flat=True)
        return super().partition_sql(expression)

    def _partition_by_sql(
        self, expression: exp.PartitionByRangeProperty | exp.PartitionByListProperty, kind: str
    ) -> str:
        partitions = self.expressions(expression, key="partition_expressions", flat=True)
        create = self.expressions(expression, key="create_expressions", flat=True)
        return f"PARTITION BY {kind} ({partitions}) ({create})"

    def partitionbyrangeproperty_sql(self, expression: exp.PartitionByRangeProperty) -> str:
        return self._partition_by_sql(expression, "RANGE")

    def partitionbylistproperty_sql(self, expression: exp.PartitionByListProperty) -> str:
        return self._partition_by_sql(expression, "LIST")

    def partitionlist_sql(self, expression: exp.PartitionList) -> str:
        name = self.sql(expression, "this")
        values = self.expressions(expression, flat=True)
        return f"PARTITION {name} VALUES IN ({values})"

    def partitionrange_sql(self, expression: exp.PartitionRange) -> str:
        name = self.sql(expression, "this")
        values = self.expressions(expression, flat=True)
        return f"PARTITION {name} VALUES LESS THAN ({values})"



from __future__ import annotations


from sqlglot import exp
from sqlglot.parsers.postgres import PostgresParser
from sqlglot.tokens import TokenType


class RisingWaveParser(PostgresParser):
    WRAPPED_TRANSFORM_COLUMN_CONSTRAINT = False

    PROPERTY_PARSERS = {
        **PostgresParser.PROPERTY_PARSERS,
        "ENCODE": lambda self: self._parse_encode_property(),
        "INCLUDE": lambda self: self._parse_include_property(),
        "KEY": lambda self: self._parse_encode_property(key=True),
    }

    CONSTRAINT_PARSERS = {
        **PostgresParser.CONSTRAINT_PARSERS,
        "WATERMARK": lambda self: self.expression(
            exp.WatermarkColumnConstraint(
                this=self._match(TokenType.FOR) and self._parse_column(),
                expression=self._match(TokenType.ALIAS) and self._parse_disjunction(),
            )
        ),
    }

    SCHEMA_UNNAMED_CONSTRAINTS = {
        *PostgresParser.SCHEMA_UNNAMED_CONSTRAINTS,
        "WATERMARK",
    }

    def _parse_table_hints(self) -> list[exp.Expr] | None:
        # There is no hint in risingwave.
        # Do nothing here to avoid WITH keywords conflict in CREATE SINK statement.
        return None

    def _parse_include_property(self) -> exp.Expr | None:
        header: exp.Expr | None = None
        coldef: exp.Expr | None = None

        this = self._parse_var_or_string()

        if not self._match(TokenType.ALIAS):
            header = self._parse_field()
            if header:
                coldef = self.expression(exp.ColumnDef(this=header, kind=self._parse_types()))

        self._match(TokenType.ALIAS)
        alias = self._parse_id_var(tokens=self.ALIAS_TOKENS)

        return self.expression(exp.IncludeProperty(this=this, alias=alias, column_def=coldef))

    def _parse_encode_property(self, key: bool | None = None) -> exp.EncodeProperty:
        self._match_text_seq("ENCODE")
        this = self._parse_var_or_string()

        if self._match(TokenType.L_PAREN, advance=False):
            properties = self.expression(
                exp.Properties(expressions=self._parse_wrapped_properties())
            )
        else:
            properties = None

        return self.expression(exp.EncodeProperty(this=this, properties=properties, key=key))



from __future__ import annotations

import typing as t

from sqlglot import exp, generator, transforms
from sqlglot.dialects.dialect import (
    bool_xor_sql,
    bracket_to_element_at_sql,
    datestrtodate_sql,
    encode_decode_sql,
    if_sql,
    left_to_substring_sql,
    no_ilike_sql,
    no_pivot_sql,
    no_timestamp_sql,
    regexp_extract_sql,
    rename_func,
    right_to_substring_sql,
    strposition_sql,
    struct_extract_sql,
    timestamptrunc_sql,
    timestrtotime_sql,
    ts_or_ds_add_cast,
    unit_to_str,
    week_unit_to_dow,
    sequence_sql,
    explode_to_unnest_sql,
)
from sqlglot.dialects.hive import Hive
from sqlglot.generator import unsupported_args
from sqlglot.optimizer.scope import find_all_in_scope
from sqlglot.transforms import unqualify_columns

DATE_ADD_OR_SUB = t.Union[exp.DateAdd, exp.TimestampAdd, exp.DateSub]


def _sha2_digest_sql(self: PrestoGenerator, expression: exp.SHA2Digest) -> str:
    length = expression.text("length") or "256"
    if length not in ("256", "512"):
        self.unsupported(f"SHA{length} is not supported in Presto")

    this = expression.this
    if this.is_type(*exp.DataType.TEXT_TYPES):
        # the native digest takes VARBINARY, so a text-typed argument needs encoding
        this = exp.Encode(this=this, charset=exp.Literal.string("utf-8"))

    return self.func(f"SHA{length}", this)


def _initcap_sql(self: PrestoGenerator, expression: exp.Initcap) -> str:
    delimiters = expression.expression
    if delimiters and not (
        delimiters.is_string and delimiters.this == self.dialect.INITCAP_DEFAULT_DELIMITER_CHARS
    ):
        self.unsupported("INITCAP does not support custom delimiters")

    regex = r"(\w)(\w*)"
    return f"REGEXP_REPLACE({self.sql(expression, 'this')}, '{regex}', x -> UPPER(x[1]) || LOWER(x[2]))"


def _no_sort_array(self: PrestoGenerator, expression: exp.SortArray) -> str:
    if expression.args.get("asc") == exp.false():
        comparator = "(a, b) -> CASE WHEN a < b THEN 1 WHEN a > b THEN -1 ELSE 0 END"
    else:
        comparator = None
    return self.func("ARRAY_SORT", expression.this, comparator)


def _schema_sql(self: PrestoGenerator, expression: exp.Schema) -> str:
    if isinstance(expression.parent, exp.PartitionedByProperty):
        # Any columns in the ARRAY[] string literals should not be quoted
        expression.transform(lambda n: n.name if isinstance(n, exp.Identifier) else n, copy=False)

        partition_exprs = [
            self.sql(c) if isinstance(c, (exp.Func, exp.Property)) else self.sql(c, "this")
            for c in expression.expressions
        ]
        return self.sql(exp.Array(expressions=[exp.Literal.string(c) for c in partition_exprs]))

    if expression.parent:
        for schema in expression.parent.find_all(exp.Schema):
            if schema is expression:
                continue

            column_defs = schema.find_all(exp.ColumnDef)
            if column_defs and isinstance(schema.parent, exp.Property):
                expression.expressions.extend(column_defs)

    return self.schema_sql(expression)


def _quantile_sql(self: PrestoGenerator, expression: exp.Quantile) -> str:
    self.unsupported("Presto does not support exact quantiles")
    return self.func("APPROX_PERCENTILE", expression.this, expression.args.get("quantile"))


def _str_to_time_sql(
    self: PrestoGenerator, expression: exp.StrToDate | exp.StrToTime | exp.TsOrDsToDate
) -> str:
    return self.func("DATE_PARSE", expression.this, self.format_time(expression))


def _ts_or_ds_to_date_sql(self: PrestoGenerator, expression: exp.TsOrDsToDate) -> str:
    time_format = self.format_time(expression)
    dialect_class = type(self.dialect)
    if time_format and time_format not in (dialect_class.TIME_FORMAT, dialect_class.DATE_FORMAT):
        return self.sql(exp.cast(_str_to_time_sql(self, expression), exp.DType.DATE))
    return self.sql(exp.cast(exp.cast(expression.this, exp.DType.TIMESTAMP), exp.DType.DATE))


def _ts_or_ds_add_sql(self: PrestoGenerator, expression: exp.TsOrDsAdd) -> str:
    expression = ts_or_ds_add_cast(expression)
    unit = unit_to_str(expression)
    return self.func("DATE_ADD", unit, expression.expression, expression.this)


def _ts_or_ds_diff_sql(self: PrestoGenerator, expression: exp.TsOrDsDiff) -> str:
    this = exp.cast(expression.this, exp.DType.TIMESTAMP)
    expr = exp.cast(expression.expression, exp.DType.TIMESTAMP)
    unit = unit_to_str(expression)
    return self.func("DATE_DIFF", unit, expr, this)


def _first_last_sql(self: PrestoGenerator, expression: exp.Func) -> str:
    """
    Trino doesn't support FIRST / LAST as functions, but they're valid in the context
    of MATCH_RECOGNIZE, so we need to preserve them in that case. In all other cases
    they're converted into an ARBITRARY call.

    Reference: https://trino.io/docs/current/sql/match-recognize.html#logical-navigation-functions
    """
    if isinstance(expression.find_ancestor(exp.MatchRecognize, exp.Select), exp.MatchRecognize):
        return self.function_fallback_sql(expression)

    return rename_func("ARBITRARY")(self, expression)


def _unix_to_time_sql(self: PrestoGenerator, expression: exp.UnixToTime) -> str:
    scale = expression.args.get("scale")
    timestamp = self.sql(expression, "this")
    if scale in (None, exp.UnixToTime.SECONDS):
        return rename_func("FROM_UNIXTIME")(self, expression)

    return f"FROM_UNIXTIME(CAST({timestamp} AS DOUBLE) / POW(10, {scale}))"


def _to_int(self: PrestoGenerator, expression: exp.Expr) -> exp.Expr:
    if not expression.type:
        from sqlglot.optimizer.annotate_types import annotate_types

        annotate_types(expression, dialect=self.dialect)
    if expression.type and expression.type.this not in exp.DataType.INTEGER_TYPES:
        return exp.cast(expression, to=exp.DType.BIGINT)
    return expression


def _date_delta_sql(
    name: str, negate_interval: bool = False
) -> t.Callable[[PrestoGenerator, DATE_ADD_OR_SUB], str]:
    def _delta_sql(self: PrestoGenerator, expression: DATE_ADD_OR_SUB) -> str:
        interval = _to_int(self, expression.expression)
        return self.func(
            name,
            unit_to_str(expression),
            interval * (-1) if negate_interval else interval,
            expression.this,
        )

    return _delta_sql


def _date_diff_sql(
    self: PrestoGenerator, expression: exp.DateDiff | exp.DatetimeDiff | exp.TimestampDiff
) -> str:
    # Presto/Trino only expose date_diff(unit, ts1, ts2); it returns ts2 - ts1, so the
    # operands are emitted as (expression, this) to preserve `this - expression` semantics.
    this: exp.Expr = expression.this
    expr: exp.Expr = expression.expression
    unit = unit_to_str(expression)

    # DATE_DIFF counts complete units between its operands, whereas dialects that set
    # date_part_boundary count unit boundary crossings, so the operands are truncated
    # down to the unit to make the two coincide
    if unit and expression.args.get("date_part_boundary"):
        raw_unit = expression.args.get("unit")
        dow = week_unit_to_dow(raw_unit)

        if dow is not None:
            unit = exp.Literal.string("WEEK")

            # DATE_TRUNC('WEEK', ...) is Monday-based; shifting both operands by the same
            # delta realigns it to the requested week start without changing the diff
            shift_days = 1 if dow == 7 else 1 - dow
            if shift_days:
                delta = exp.Interval(this=exp.Literal.string(str(shift_days)), unit=exp.var("DAY"))
                this = exp.Add(this=this, expression=delta)
                expr = exp.Add(this=expr, expression=delta.copy())

        if not isinstance(raw_unit, exp.WeekStart) or dow is not None:
            this = exp.DateTrunc(unit=unit.copy(), this=this)
            expr = exp.DateTrunc(unit=unit.copy(), this=expr)

    return self.func("DATE_DIFF", unit, expr, this)


def _explode_to_unnest_sql(self: PrestoGenerator, expression: exp.Lateral) -> str:
    explode = expression.this
    if isinstance(explode, exp.Explode):
        exploded_type = explode.this.type
        alias = expression.args.get("alias")

        # This attempts a best-effort transpilation of LATERAL VIEW EXPLODE on a struct array
        if (
            isinstance(alias, exp.TableAlias)
            and isinstance(exploded_type, exp.DataType)
            and exploded_type.is_type(exp.DType.ARRAY)
            and exploded_type.expressions
            and exploded_type.expressions[0].is_type(exp.DType.STRUCT)
        ):
            # When unnesting a ROW in Presto, it produces N columns, so we need to fix the alias
            alias.set("columns", [c.this.copy() for c in exploded_type.expressions[0].expressions])
    elif isinstance(explode, exp.Inline):
        explode.replace(exp.Explode(this=explode.this.copy()))

    return explode_to_unnest_sql(self, expression)


def amend_exploded_column_table(expression: exp.Expr) -> exp.Expr:
    # We check for expression.type because the columns can be amended only if types were inferred
    if isinstance(expression, exp.Select) and expression.type:
        for lateral in expression.args.get("laterals") or []:
            alias = lateral.args.get("alias")
            if (
                not isinstance(lateral.this, exp.Explode)
                or not isinstance(alias, exp.TableAlias)
                or len(alias.columns) != 1
            ):
                continue

            new_table = alias.this
            old_table = alias.columns[0].name.lower()

            # When transpiling a LATERAL VIEW EXPLODE Spark query, the exploded fields may be qualified
            # with the struct column, resulting in invalid Presto references that need to be amended
            for column in find_all_in_scope(expression, exp.Column):
                if column.db.lower() == old_table:
                    column.set("table", column.args["db"].pop())
                elif column.table.lower() == old_table:
                    column.set("table", new_table.copy())
                elif column.name.lower() == old_table and isinstance(column.parent, exp.Dot):
                    column.parent.replace(exp.column(column.parent.expression, table=new_table))

    return expression


class PrestoGenerator(generator.Generator):
    SELECT_KINDS: tuple[str, ...] = ()
    SUPPORTS_DECODE_CASE = False

    AFTER_HAVING_MODIFIER_TRANSFORMS = generator.AFTER_HAVING_MODIFIER_TRANSFORMS

    INTERVAL_ALLOWS_PLURAL_FORM = False
    JOIN_HINTS = False
    TABLE_HINTS = False
    QUERY_HINTS = False
    IS_BOOL_ALLOWED = False
    TZ_TO_WITH_TIME_ZONE = True
    NVL2_SUPPORTED = False
    STRUCT_DELIMITER = ("(", ")")
    LIMIT_ONLY_LITERALS = True
    SUPPORTS_SINGLE_ARG_CONCAT = False
    LIKE_PROPERTY_INSIDE_SCHEMA = True
    MULTI_ARG_DISTINCT = False
    SUPPORTS_TO_NUMBER = False
    HEX_FUNC = "TO_HEX"
    PARSE_JSON_NAME: str | None = "JSON_PARSE"
    PAD_FILL_PATTERN_IS_REQUIRED = True
    EXCEPT_INTERSECT_SUPPORT_ALL_CLAUSE = False
    SUPPORTS_MEDIAN = False
    ARRAY_SIZE_NAME = "CARDINALITY"

    PROPERTIES_LOCATION = {
        **generator.Generator.PROPERTIES_LOCATION,
        exp.LocationProperty: exp.Properties.Location.UNSUPPORTED,
        exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    }

    TYPE_MAPPING = {
        **generator.Generator.TYPE_MAPPING,
        exp.DType.BINARY: "VARBINARY",
        exp.DType.BIT: "BOOLEAN",
        exp.DType.DATETIME: "TIMESTAMP",
        exp.DType.DATETIME64: "TIMESTAMP",
        exp.DType.FLOAT: "REAL",
        exp.DType.HLLSKETCH: "HYPERLOGLOG",
        exp.DType.INT: "INTEGER",
        exp.DType.STRUCT: "ROW",
        exp.DType.TEXT: "VARCHAR",
        exp.DType.TIMESTAMPTZ: "TIMESTAMP",
        exp.DType.TIMESTAMPNTZ: "TIMESTAMP",
        exp.DType.TIMETZ: "TIME",
    }

    TRANSFORMS = {
        **generator.Generator.TRANSFORMS,
        exp.AnyValue: rename_func("ARBITRARY"),
        exp.ApproxQuantile: lambda self, e: self.func(
            "APPROX_PERCENTILE",
            e.this,
            e.args.get("weight"),
            e.args.get("quantile"),
            e.args.get("accuracy"),
        ),
        exp.ArgMax: rename_func("MAX_BY"),
        exp.ArgMin: rename_func("MIN_BY"),
        exp.Array: transforms.preprocess(
            [transforms.inherit_struct_field_names],
            generator=lambda self, e: f"ARRAY[{self.expressions(e, flat=True)}]",
        ),
        exp.ArrayAny: rename_func("ANY_MATCH"),
        exp.ArrayConcat: rename_func("CONCAT"),
        exp.ArrayContains: rename_func("CONTAINS"),
        exp.ArrayToString: rename_func("ARRAY_JOIN"),
        exp.ArrayUniqueAgg: rename_func("SET_AGG"),
        exp.ArraySlice: rename_func("SLICE"),
        exp.AtTimeZone: rename_func("AT_TIMEZONE"),
        exp.BitwiseAnd: lambda self, e: self.func("BITWISE_AND", e.this, e.expression),
        exp.BitwiseLeftShift: rename_func("BITWISE_LEFT_SHIFT"),
        exp.BitwiseNot: lambda self, e: self.func("BITWISE_NOT", e.this),
        exp.BitwiseOr: lambda self, e: self.func("BITWISE_OR", e.this, e.expression),
        exp.BitwiseRightShift: rename_func("BITWISE_RIGHT_SHIFT"),
        exp.BitwiseXor: lambda self, e: self.func("BITWISE_XOR", e.this, e.expression),
        exp.Cast: transforms.preprocess([transforms.epoch_cast_to_ts]),
        exp.CurrentTime: lambda *_: "CURRENT_TIME",
        exp.CurrentTimestamp: lambda *_: "CURRENT_TIMESTAMP",
        exp.CurrentUser: lambda *_: "CURRENT_USER",
        exp.DateAdd: _date_delta_sql("DATE_ADD"),
        exp.DateDiff: _date_diff_sql,
        exp.DatetimeDiff: _date_diff_sql,
        exp.TimestampDiff: _date_diff_sql,
        exp.DateStrToDate: datestrtodate_sql,
        exp.DateToDi: lambda self, e: (
            f"CAST(DATE_FORMAT({self.sql(e, 'this')}, {type(self.dialect).DATEINT_FORMAT}) AS INT)"
        ),
        exp.DateSub: _date_delta_sql("DATE_ADD", negate_interval=True),
        exp.DayOfWeek: lambda self, e: f"(({self.func('DAY_OF_WEEK', e.this)} % 7) + 1)",
        exp.DayOfWeekIso: rename_func("DAY_OF_WEEK"),
        exp.Decode: lambda self, e: encode_decode_sql(self, e, "FROM_UTF8"),
        exp.DiToDate: lambda self, e: (
            f"CAST(DATE_PARSE(CAST({self.sql(e, 'this')} AS VARCHAR), {type(self.dialect).DATEINT_FORMAT}) AS DATE)"
        ),
        exp.Encode: lambda self, e: encode_decode_sql(self, e, "TO_UTF8"),
        exp.FileFormatProperty: lambda self, e: f"format={self.sql(exp.Literal.string(e.name))}",
        exp.First: _first_last_sql,
        exp.FromISO8601Date: rename_func("FROM_ISO8601_DATE"),
        exp.FromISO8601Timestamp: rename_func("FROM_ISO8601_TIMESTAMP"),
        exp.FromTimeZone: lambda self, e: (
            f"WITH_TIMEZONE({self.sql(e, 'this')}, {self.sql(e, 'zone')}) AT TIME ZONE 'UTC'"
        ),
        exp.GenerateSeries: sequence_sql,
        exp.GenerateDateArray: sequence_sql,
        exp.If: if_sql(),
        exp.ILike: no_ilike_sql,
        exp.Initcap: _initcap_sql,
        exp.Last: _first_last_sql,
        exp.LastDay: lambda self, e: self.func("LAST_DAY_OF_MONTH", e.this),
        exp.Lateral: _explode_to_unnest_sql,
        exp.Left: left_to_substring_sql,
        exp.Levenshtein: unsupported_args("ins_cost", "del_cost", "sub_cost", "max_dist")(
            rename_func("LEVENSHTEIN_DISTANCE")
        ),
        exp.LogicalAnd: rename_func("BOOL_AND"),
        exp.LogicalOr: rename_func("BOOL_OR"),
        exp.Pivot: no_pivot_sql,
        exp.Quantile: _quantile_sql,
        exp.RegexpExtract: regexp_extract_sql,
        exp.RegexpExtractAll: regexp_extract_sql,
        exp.Right: right_to_substring_sql,
        exp.Schema: _schema_sql,
        exp.SchemaCommentProperty: lambda self, e: self.naked_property(e),
        exp.Select: transforms.preprocess(
            [
                transforms.eliminate_window_clause,
                transforms.eliminate_qualify,
                transforms.eliminate_distinct_on,
                transforms.explode_projection_to_unnest(1),
                transforms.eliminate_semi_and_anti_joins,
                amend_exploded_column_table,
            ]
        ),
        exp.SortArray: _no_sort_array,
        exp.SqlSecurityProperty: lambda self, e: f"SECURITY {self.sql(e.this)}",
        exp.StrPosition: lambda self, e: strposition_sql(self, e, supports_occurrence=True),
        exp.StrToDate: lambda self, e: f"CAST({_str_to_time_sql(self, e)} AS DATE)",
        exp.StrToMap: rename_func("SPLIT_TO_MAP"),
        exp.StrToTime: _str_to_time_sql,
        exp.StructExtract: struct_extract_sql,
        exp.Table: transforms.preprocess([transforms.unnest_generate_series]),
        exp.Timestamp: no_timestamp_sql,
        exp.TimestampAdd: _date_delta_sql("DATE_ADD"),
        exp.TimestampTrunc: timestamptrunc_sql(),
        exp.TimeStrToDate: timestrtotime_sql,
        exp.TimeStrToTime: timestrtotime_sql,
        exp.TimeStrToUnix: lambda self, e: self.func(
            "TO_UNIXTIME", self.func("DATE_PARSE", e.this, type(self.dialect).TIME_FORMAT)
        ),
        exp.TimeToStr: lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e)),
        exp.TimeToUnix: rename_func("TO_UNIXTIME"),
        exp.ToChar: lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e)),
        exp.TryCast: transforms.preprocess([transforms.epoch_cast_to_ts]),
        exp.TsOrDiToDi: lambda self, e: (
            f"CAST(SUBSTR(REPLACE(CAST({self.sql(e, 'this')} AS VARCHAR), '-', ''), 1, 8) AS INT)"
        ),
        exp.TsOrDsAdd: _ts_or_ds_add_sql,
        exp.TsOrDsDiff: _ts_or_ds_diff_sql,
        exp.TsOrDsToDate: _ts_or_ds_to_date_sql,
        exp.Unhex: rename_func("FROM_HEX"),
        exp.UnixToStr: lambda self, e: (
            f"DATE_FORMAT(FROM_UNIXTIME({self.sql(e, 'this')}), {self.format_time(e)})"
        ),
        exp.UnixToTime: _unix_to_time_sql,
        exp.UnixToTimeStr: lambda self, e: f"CAST(FROM_UNIXTIME({self.sql(e, 'this')}) AS VARCHAR)",
        exp.VariancePop: rename_func("VAR_POP"),
        exp.With: transforms.preprocess([transforms.add_recursive_cte_column_names]),
        exp.WithinGroup: transforms.preprocess([transforms.remove_within_group_for_percentiles]),
        # Note: Presto's TRUNCATE always returns DOUBLE, even with decimals=0, whereas
        # most dialects return INT (SQLite also returns REAL, see sqlite.py). This creates
        # a bidirectional transpilation gap: Presto→Other may change float division to int
        # division, and vice versa. Modeling precisely would require exp.FloatTrunc or
        # similar, deemed overengineering for this subtle semantic difference.
        exp.Trunc: rename_func("TRUNCATE"),
        exp.Xor: bool_xor_sql,
        exp.MD5Digest: rename_func("MD5"),
        exp.SHA: rename_func("SHA1"),
        exp.SHA1Digest: rename_func("SHA1"),
        exp.SHA2Digest: _sha2_digest_sql,
        exp.Substring: rename_func("SUBSTR"),
    }

    RESERVED_KEYWORDS = {
        "alter",
        "and",
        "as",
        "between",
        "by",
        "case",
        "cast",
        "constraint",
        "create",
        "cross",
        "current_time",
        "current_timestamp",
        "deallocate",
        "delete",
        "describe",
        "distinct",
        "drop",
        "else",
        "end",
        "escape",
        "except",
        "execute",
        "exists",
        "extract",
        "false",
        "for",
        "from",
        "full",
        "group",
        "having",
        "in",
        "inner",
        "insert",
        "intersect",
        "into",
        "is",
        "join",
        "left",
        "like",
        "natural",
        "not",
        "null",
        "on",
        "or",
        "order",
        "outer",
        "prepare",
        "right",
        "select",
        "table",
        "then",
        "true",
        "union",
        "using",
        "values",
        "when",
        "where",
        "with",
    }

    def extract_sql(self, expression: exp.Extract) -> str:
        date_part = expression.name

        if not date_part.startswith("EPOCH"):
            return super().extract_sql(expression)

        if date_part == "EPOCH_MILLISECOND":
            scale = 10**3
        elif date_part == "EPOCH_MICROSECOND":
            scale = 10**6
        elif date_part == "EPOCH_NANOSECOND":
            scale = 10**9
        else:
            scale = None

        value = expression.expression

        ts = exp.cast(value, to=exp.DType.TIMESTAMP.into_expr())
        to_unix: exp.Expr = exp.TimeToUnix(this=ts)

        if scale:
            to_unix = exp.Mul(this=to_unix, expression=exp.Literal.number(scale))

        return self.sql(to_unix)

    def jsonformat_sql(self, expression: exp.JSONFormat) -> str:
        this = expression.this
        is_json = expression.args.get("is_json")

        if this and not (is_json or this.type):
            from sqlglot.optimizer.annotate_types import annotate_types

            this = annotate_types(this, dialect=self.dialect)

        if not (is_json or this.is_type(exp.DType.JSON)):
            this.replace(exp.cast(this, exp.DType.JSON))

        return self.function_fallback_sql(expression)

    def md5_sql(self, expression: exp.MD5) -> str:
        this = expression.this

        if not this.type:
            from sqlglot.optimizer.annotate_types import annotate_types

            this = annotate_types(this, dialect=self.dialect)

        if this.is_type(*exp.DataType.TEXT_TYPES):
            this = exp.Encode(this=this, charset=exp.Literal.string("utf-8"))

        return self.func("LOWER", self.func("TO_HEX", self.func("MD5", self.sql(this))))

    def sha2_sql(self, expression: exp.SHA2) -> str:
        length = expression.text("length") or "256"
        if length not in ("256", "512"):
            self.unsupported(f"SHA{length} is not supported in Presto")

        this = expression.this

        if this.is_type(*exp.DataType.TEXT_TYPES):
            this = exp.Encode(this=this, charset=exp.Literal.string("utf-8"))

        return self.func("LOWER", self.func("TO_HEX", self.func(f"SHA{length}", self.sql(this))))

    def strtounix_sql(self, expression: exp.StrToUnix) -> str:
        # Since `TO_UNIXTIME` requires a `TIMESTAMP`, we need to parse the argument into one.
        # To do this, we first try to `DATE_PARSE` it, but since this can fail when there's a
        # timezone involved, we wrap it in a `TRY` call and use `PARSE_DATETIME` as a fallback,
        # which seems to be using the same time mapping as Hive, as per:
        # https://joda-time.sourceforge.net/apidocs/org/joda/time/format/DateTimeFormat.html
        this = expression.this
        value_as_text = exp.cast(this, exp.DType.TEXT)
        value_as_timestamp = exp.cast(this, exp.DType.TIMESTAMP) if this.is_string else this

        parse_without_tz = self.func("DATE_PARSE", value_as_text, self.format_time(expression))

        formatted_value = self.func("DATE_FORMAT", value_as_timestamp, self.format_time(expression))
        parse_with_tz = self.func(
            "PARSE_DATETIME",
            formatted_value,
            self.format_time(expression, Hive.INVERSE_TIME_MAPPING, Hive.INVERSE_TIME_TRIE),
        )
        coalesced = self.func("COALESCE", self.func("TRY", parse_without_tz), parse_with_tz)
        return self.func("TO_UNIXTIME", coalesced)

    def bracket_sql(self, expression: exp.Bracket) -> str:
        if expression.args.get("safe"):
            return bracket_to_element_at_sql(self, expression)
        return super().bracket_sql(expression)

    def struct_sql(self, expression: exp.Struct) -> str:
        if not expression.type:
            from sqlglot.optimizer.annotate_types import annotate_types

            annotate_types(expression, dialect=self.dialect)

        values: list[str] = []
        schema: list[str] = []
        unknown_type = False

        for e in expression.expressions:
            if isinstance(e, exp.PropertyEQ):
                if e.type and e.type.is_type(exp.DType.UNKNOWN):
                    unknown_type = True
                else:
                    schema.append(f"{self.sql(e, 'this')} {self.sql(e.type)}")
                values.append(self.sql(e, "expression"))
            else:
                values.append(self.sql(e))

        size = len(expression.expressions)

        if not size or len(schema) != size:
            if unknown_type:
                self.unsupported(
                    "Cannot convert untyped key-value definitions (try annotate_types)."
                )
            return self.func("ROW", *values)
        return f"CAST(ROW({', '.join(values)}) AS ROW({', '.join(schema)}))"

    def interval_sql(self, expression: exp.Interval) -> str:
        if expression.this and expression.text("unit").upper().startswith("WEEK"):
            return f"({expression.this.name} * INTERVAL '7' DAY)"
        return super().interval_sql(expression)

    def transaction_sql(self, expression: exp.Transaction) -> str:
        modes = expression.args.get("modes")
        modes = f" {', '.join(modes)}" if modes else ""
        return f"START TRANSACTION{modes}"

    def offset_limit_modifiers(
        self, expression: exp.Expr, fetch: bool, limit: exp.Fetch | exp.Limit | None
    ) -> list[str]:
        return [
            self.sql(expression, "offset"),
            self.sql(limit),
        ]

    def create_sql(self, expression: exp.Create) -> str:
        """
        Presto doesn't support CREATE VIEW with expressions (ex: `CREATE VIEW x (cola)` then `(cola)` is the expression),
        so we need to remove them
        """
        kind = expression.args["kind"]
        schema = expression.this
        if kind == "VIEW" and schema.expressions:
            expression.this.set("expressions", None)
        return super().create_sql(expression)

    def delete_sql(self, expression: exp.Delete) -> str:
        """
        Presto only supports DELETE FROM for a single table without an alias, so we need
        to remove the unnecessary parts. If the original DELETE statement contains more
        than one table to be deleted, we can't safely map it 1-1 to a Presto statement.
        """
        tables = expression.args.get("tables") or [expression.this]
        if len(tables) > 1:
            return super().delete_sql(expression)

        table = tables[0]
        expression.set("this", table)
        expression.set("tables", None)

        if isinstance(table, exp.Table):
            table_alias = table.args.get("alias")
            if table_alias:
                table_alias.pop()
                expression = t.cast(exp.Delete, expression.transform(unqualify_columns))

        return super().delete_sql(expression)

    def jsonextract_sql(self, expression: exp.JSONExtract) -> str:
        is_json_extract = self.dialect.settings.get("variant_extract_is_json_extract", True)

        # Generate JSON_EXTRACT unless the user has configured that a Snowflake / Databricks
        # VARIANT extract (e.g. col:x.y) should map to dot notation (i.e ROW access) in Presto/Trino
        if not expression.args.get("variant_extract") or is_json_extract:
            return self.func(
                "JSON_EXTRACT", expression.this, expression.expression, *expression.expressions
            )

        this = self.sql(expression, "this")

        # Convert the JSONPath extraction `JSON_EXTRACT(col, '$.x.y) to a ROW access col.x.y
        segments = []
        for path_key in expression.expression.expressions[1:]:
            if not isinstance(path_key, exp.JSONPathKey):
                # Cannot transpile subscripts, wildcards etc to dot notation
                self.unsupported(f"Cannot transpile JSONPath segment '{path_key}' to ROW access")
                continue
            key = path_key.this
            if not exp.SAFE_IDENTIFIER_RE.match(key):
                key = f'"{key}"'
            segments.append(f".{key}")

        expr = "".join(segments)

        return f"{this}{expr}"

    def groupconcat_sql(self, expression: exp.GroupConcat) -> str:
        return self.func(
            "ARRAY_JOIN",
            self.func("ARRAY_AGG", expression.this),
            expression.args.get("separator"),
        )



"""sqlglot expressions properties."""

from __future__ import annotations

import typing as t
from enum import auto

from sqlglot.expressions.core import ColumnConstraintKind, Expression, Literal, convert
from sqlglot.helper import AutoName


class Property(Expression):
    arg_types = {"this": True, "value": True}


class GrantPrivilege(Expression):
    arg_types = {"this": True, "expressions": False}


class GrantPrincipal(Expression):
    arg_types = {"this": True, "kind": False}


class AllowedValuesProperty(Expression):
    arg_types = {"expressions": True}


class AlgorithmProperty(Property):
    arg_types = {"this": True}


class ApiProperty(Property):
    arg_types = {}


class ApplicationProperty(Property):
    arg_types = {}


class AutoIncrementProperty(Property):
    arg_types = {"this": True}


class AutoRefreshProperty(Property):
    arg_types = {
        "this": False,
        "cadence": False,
        "offset": False,
        "randomize": False,
        "expressions": False,
        "settings": False,
        "append": False,
    }


class BackupProperty(Property):
    arg_types = {"this": True}


class BuildProperty(Property):
    arg_types = {"this": True}


class BlockCompressionProperty(Property):
    arg_types = {
        "autotemp": False,
        "always": False,
        "default": False,
        "manual": False,
        "never": False,
    }


class CalledOnNullInputProperty(Property):
    arg_types = {}


class CatalogProperty(Property):
    arg_types = {}


class CharacterSetProperty(Property):
    arg_types = {"this": True, "default": True}


class ChecksumProperty(Property):
    arg_types = {"on": False, "default": False}


class ClusterProperty(Property):
    arg_types = {"this": False, "expressions": False}


class CollateProperty(Property):
    arg_types = {"this": True, "default": False}


class ComputeProperty(Property):
    arg_types = {}


class CopyGrantsProperty(Property):
    arg_types = {}


class DataBlocksizeProperty(Property):
    arg_types = {
        "size": False,
        "units": False,
        "minimum": False,
        "maximum": False,
        "default": False,
    }


class DataDeletionProperty(Property):
    arg_types = {"on": True, "filter_column": False, "retention_period": False}


class DatabaseProperty(Property):
    arg_types = {}


class DefinerProperty(Property):
    arg_types = {"this": True}


class DistKeyProperty(Property):
    arg_types = {"this": True}


class DistributedByProperty(Property):
    arg_types = {"expressions": False, "kind": True, "buckets": False, "order": False}


class DistStyleProperty(Property):
    arg_types = {"this": True}


class DuplicateKeyProperty(Property):
    arg_types = {"expressions": True}


class EngineProperty(Property):
    arg_types = {"this": True}


class UuidProperty(Property):
    arg_types = {"this": True}


class HeapProperty(Property):
    arg_types = {}


class HybridProperty(Property):
    arg_types = {}


class HandlerProperty(Property):
    arg_types = {"this": True}


class ParameterStyleProperty(Property):
    arg_types = {"this": True}


class ToTableProperty(Property):
    arg_types = {"this": True}


class ExecuteAsProperty(Property):
    arg_types = {"this": True}


class ExternalProperty(Property):
    arg_types = {"this": False}


class FallbackProperty(Property):
    arg_types = {"no": True, "protection": False}


class FileFormatProperty(Property):
    arg_types = {"this": False, "expressions": False, "hive_format": False}


class CredentialsProperty(Property):
    arg_types = {"expressions": True}


class FreespaceProperty(Property):
    arg_types = {"this": True, "percent": False}


class GlobalProperty(Property):
    arg_types = {}


class IcebergProperty(Property):
    arg_types = {}


class InheritsProperty(Property):
    arg_types = {"expressions": True}


class InputModelProperty(Property):
    arg_types = {"this": True}


class OutputModelProperty(Property):
    arg_types = {"this": True}


class IsolatedLoadingProperty(Property):
    arg_types = {"no": False, "concurrent": False, "target": False}


class JournalProperty(Property):
    arg_types = {
        "no": False,
        "dual": False,
        "before": False,
        "local": False,
        "after": False,
    }


class LanguageProperty(Property):
    arg_types = {"this": True}


class EnviromentProperty(Property):
    arg_types = {"expressions": True}


class ClusteredByProperty(Property):
    arg_types = {"expressions": True, "sorted_by": False, "buckets": True}


class DictProperty(Property):
    arg_types = {"this": True, "kind": True, "settings": False}


class DictSubProperty(Property):
    pass


class DictRange(Property):
    arg_types = {"this": True, "min": True, "max": True}


class DynamicProperty(Property):
    arg_types = {}


class OnCluster(Property):
    arg_types = {"this": True}


class EmptyProperty(Property):
    arg_types = {}


class LikeProperty(Property):
    arg_types = {"this": True, "expressions": False}


class LocationProperty(Property):
    arg_types = {"this": True}


class LockProperty(Property):
    arg_types = {"this": True}


class LockingProperty(Property):
    arg_types = {
        "this": False,
        "kind": True,
        "for_or_in": False,
        "lock_type": True,
        "override": False,
    }


class LogProperty(Property):
    arg_types = {"no": True}


class MaskingProperty(Property):
    arg_types = {}


class MaterializedProperty(Property):
    arg_types = {"this": False}


class MergeBlockRatioProperty(Property):
    arg_types = {"this": False, "no": False, "default": False, "percent": False}


class ModuleProperty(Property):
    arg_types = {"this": True, "expressions": False}


class NetworkProperty(Property):
    arg_types = {}


class NoPrimaryIndexProperty(Property):
    arg_types = {}


class OnProperty(Property):
    arg_types = {"this": True}


class OnCommitProperty(Property):
    arg_types = {"delete": False}


class PartitionedByProperty(Property):
    arg_types = {"this": True}


class PartitionedByBucket(Property):
    arg_types = {"this": True, "expression": True}


class PartitionByTruncate(Property):
    arg_types = {"this": True, "expression": True}


class PartitionByRangeProperty(Property):
    arg_types = {"partition_expressions": True, "create_expressions": True}


class PartitionByRangePropertyDynamic(Expression):
    arg_types = {"this": False, "start": True, "end": True, "every": True}


class RollupProperty(Property):
    arg_types = {"expressions": True}


class RollupIndex(Expression):
    arg_types = {"this": True, "expressions": True, "from_index": False, "properties": False}


class RowAccessProperty(Property):
    arg_types = {"this": False, "expressions": False}


class PartitionByListProperty(Property):
    arg_types = {"partition_expressions": True, "create_expressions": True}


class PartitionList(Expression):
    arg_types = {"this": True, "expressions": True}


class RefreshTriggerProperty(Property):
    arg_types = {
        "method": False,
        "kind": False,
        "every": False,
        "unit": False,
        "starts": False,
    }


class UniqueKeyProperty(Property):
    arg_types = {"expressions": True}


class PartitionBoundSpec(Expression):
    # this -> IN / MODULUS, expression -> REMAINDER, from_expressions -> FROM (...), to_expressions -> TO (...)
    arg_types = {
        "this": False,
        "expression": False,
        "from_expressions": False,
        "to_expressions": False,
    }


class PartitionedOfProperty(Property):
    # this -> parent_table (schema), expression -> FOR VALUES ... / DEFAULT
    arg_types = {"this": True, "expression": True}


class StreamingTableProperty(Property):
    arg_types = {}


class RemoteWithConnectionModelProperty(Property):
    arg_types = {"this": True}


class ReturnsProperty(Property):
    arg_types = {"this": False, "is_table": False, "table": False, "null": False}


class StrictProperty(Property):
    arg_types = {}


class RowFormatProperty(Property):
    arg_types = {"this": True}


class RowFormatDelimitedProperty(Property):
    # https://cwiki.apache.org/confluence/display/hive/languagemanual+dml
    arg_types = {
        "fields": False,
        "escaped": False,
        "collection_items": False,
        "map_keys": False,
        "lines": False,
        "null": False,
        "serde": False,
    }


class RowFormatSerdeProperty(Property):
    arg_types = {"this": True, "serde_properties": False}


class QueryTransform(Expression):
    arg_types = {
        "expressions": True,
        "command_script": True,
        "schema": False,
        "row_format_before": False,
        "record_writer": False,
        "row_format_after": False,
        "record_reader": False,
    }


class SampleProperty(Property):
    arg_types = {"this": True}


class SchemaCommentProperty(Property):
    arg_types = {"this": True}


class SemanticView(Expression):
    arg_types = {
        "this": True,
        "metrics": False,
        "dimensions": False,
        "facts": False,
        "where": False,
    }


class SerdeProperties(Property):
    arg_types = {"expressions": True, "with_": False}


class SetProperty(Property):
    arg_types = {"multi": True}


class SharingProperty(Property):
    arg_types = {"this": False}


class SetConfigProperty(Property):
    arg_types = {"this": True}


class SettingsProperty(Property):
    arg_types = {"expressions": True}


class SortKeyProperty(Property):
    arg_types = {"this": True, "compound": False}


class SqlReadWriteProperty(Property):
    arg_types = {"this": True}


class SqlSecurityProperty(Property):
    arg_types = {"this": True}


class StabilityProperty(Property):
    arg_types = {"this": True}


class StorageHandlerProperty(Property):
    arg_types = {"this": True}


class UsingProperty(Property):
    # kind: JAR, FILE, or ARCHIVE; this: the resource path (string literal)
    arg_types = {"this": True, "kind": True}


class TemporaryProperty(Property):
    arg_types = {"this": False}


class VirtualProperty(Property):
    arg_types = {}


class SecureProperty(Property):
    arg_types = {}


class SecurityIntegrationProperty(Property):
    arg_types = {}


class Tags(Property, ColumnConstraintKind):
    arg_types = {"expressions": True}


class PropertiesLocation(AutoName):
    POST_CREATE = auto()
    POST_NAME = auto()
    POST_SCHEMA = auto()
    POST_WITH = auto()
    POST_ALIAS = auto()
    POST_EXPRESSION = auto()
    POST_INDEX = auto()
    UNSUPPORTED = auto()


class TransformModelProperty(Property):
    arg_types = {"expressions": True}


class TransientProperty(Property):
    arg_types = {"this": False}


class UnloggedProperty(Property):
    arg_types = {}


class UsingTemplateProperty(Property):
    arg_types = {"this": True}


class ViewAttributeProperty(Property):
    arg_types = {"this": True}


class VolatileProperty(Property):
    arg_types = {"this": False}


class WithDataProperty(Property):
    arg_types = {"no": True, "statistics": False}


class WithJournalTableProperty(Property):
    arg_types = {"this": True}


class WithSchemaBindingProperty(Property):
    arg_types = {"this": True}


class WithSystemVersioningProperty(Property):
    arg_types = {
        "on": False,
        "this": False,
        "data_consistency": False,
        "retention_period": False,
        "with_": True,
    }


class WithProcedureOptions(Property):
    arg_types = {"expressions": True}


class EncodeProperty(Property):
    arg_types = {"this": True, "properties": False, "key": False}


class IncludeProperty(Property):
    arg_types = {"this": True, "alias": False, "column_def": False}


class ForceProperty(Property):
    arg_types = {}


class Properties(Expression):
    arg_types = {"expressions": True}

    NAME_TO_PROPERTY: t.ClassVar[dict[str, type[Property]]] = {
        "ALGORITHM": AlgorithmProperty,
        "AUTO_INCREMENT": AutoIncrementProperty,
        "CHARACTER SET": CharacterSetProperty,
        "CLUSTERED_BY": ClusteredByProperty,
        "COLLATE": CollateProperty,
        "COMMENT": SchemaCommentProperty,
        "CREDENTIALS": CredentialsProperty,
        "DEFINER": DefinerProperty,
        "DISTKEY": DistKeyProperty,
        "DISTRIBUTED_BY": DistributedByProperty,
        "DISTSTYLE": DistStyleProperty,
        "ENGINE": EngineProperty,
        "EXECUTE AS": ExecuteAsProperty,
        "FORMAT": FileFormatProperty,
        "LANGUAGE": LanguageProperty,
        "LOCATION": LocationProperty,
        "LOCK": LockProperty,
        "PARTITIONED_BY": PartitionedByProperty,
        "RETURNS": ReturnsProperty,
        "ROW_FORMAT": RowFormatProperty,
        "SORTKEY": SortKeyProperty,
        "ENCODE": EncodeProperty,
        "INCLUDE": IncludeProperty,
    }

    PROPERTY_TO_NAME: t.ClassVar[dict[type[Property], str]] = {}

    # CREATE property locations
    # Form: schema specified
    #   create [POST_CREATE]
    #     table a [POST_NAME]
    #     (b int) [POST_SCHEMA]
    #     with ([POST_WITH])
    #     index (b) [POST_INDEX]
    #
    # Form: alias selection
    #   create [POST_CREATE]
    #     table a [POST_NAME]
    #     as [POST_ALIAS] (select * from b) [POST_EXPRESSION]
    #     index (c) [POST_INDEX]
    Location: t.ClassVar[type[PropertiesLocation]] = PropertiesLocation

    @classmethod
    def from_dict(cls, properties_dict: dict) -> Properties:
        expressions = []
        for key, value in properties_dict.items():
            property_cls = cls.NAME_TO_PROPERTY.get(key.upper())
            if property_cls:
                expressions.append(property_cls(this=convert(value)))
            else:
                expressions.append(Property(this=Literal.string(key), value=convert(value)))

        return cls(expressions=expressions)


# TODO (mypyc)
Properties.PROPERTY_TO_NAME = {v: k for k, v in Properties.NAME_TO_PROPERTY.items()}



from __future__ import annotations

from sqlglot import exp
from sqlglot.typing import EXPRESSION_METADATA

EXPRESSION_METADATA = {
    **EXPRESSION_METADATA,
    **{
        expr_type: {"returns": exp.DType.BINARY}
        for expr_type in {
            exp.Encode,
            exp.Unhex,
        }
    },
    **{
        expr_type: {"returns": exp.DType.DOUBLE}
        for expr_type in {
            exp.Corr,
            exp.MonthsBetween,
            exp.Sign,
        }
    },
    **{
        expr_type: {"returns": exp.DType.VARCHAR}
        for expr_type in {
            exp.AddMonths,
            exp.CurrentDatabase,
            exp.Hex,
            exp.JSONExtractScalar,
            exp.JSONFormat,
            exp.NextDay,
            exp.RegexpExtract,
            exp.RegexpReplace,
            exp.Replace,
            exp.Soundex,
        }
    },
    **{
        expr_type: {"returns": exp.DType.BIGINT}
        for expr_type in {
            exp.Factorial,
            exp.IntDiv,
            exp.StrToUnix,
        }
    },
    **{
        expr_type: {"returns": exp.DType.INT}
        for expr_type in {
            exp.ArraySize,
            exp.DenseRank,
            exp.Month,
            exp.Ntile,
            exp.Rank,
            exp.RowNumber,
            exp.Second,
            exp.Minute,
        }
    },
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_by_args(e, "this")}
        for expr_type in {
            exp.ArrayDistinct,
            exp.ArrayExcept,
            exp.First,
            exp.Last,
            exp.Negative,
            exp.Reverse,
        }
    },
    exp.ArrayIntersect: {"annotator": lambda self, e: self._annotate_by_args(e, "expressions")},
    exp.ApproxQuantile: {"annotator": lambda self, e: self._annotate_by_args(e, "quantile")},
    exp.Coalesce: {
        "annotator": lambda self, e: self._annotate_by_args(e, "this", "expressions", promote=True)
    },
    exp.Grouping: {"returns": exp.DType.BIGINT},
    exp.If: {"annotator": lambda self, e: self._annotate_by_args(e, "true", "false", promote=True)},
    exp.PercentileDisc: {"returns": exp.DType.DOUBLE},
    exp.Quantile: {"annotator": lambda self, e: self._annotate_by_args(e, "quantile")},
    exp.RegexpSplit: {"returns": exp.DataType.from_str("ARRAY<STRING>")},
    exp.StrToMap: {"returns": exp.DataType.from_str("MAP<STRING, STRING>")},
    exp.WithinGroup: {"annotator": lambda self, e: self._annotate_by_args(e, "this")},
}



from __future__ import annotations

import typing as t

from sqlglot import exp, parser
from sqlglot.dialects.dialect import build_formatted_time, build_regexp_extract
from sqlglot.helper import seq_get
from sqlglot.tokens import TokenType

if t.TYPE_CHECKING:
    from sqlglot.dialects.dialect import Dialect


def build_with_ignore_nulls(
    exp_class: type[exp.Expr],
) -> t.Callable[[list[exp.Expr]], exp.Expr]:
    def _parse(args: list[exp.Expr]) -> exp.Expr:
        this = exp_class(this=seq_get(args, 0))
        if seq_get(args, 1) == exp.true():
            return exp.IgnoreNulls(this=this)
        return this

    return _parse


def _build_to_date(args: list, dialect: Dialect) -> exp.TsOrDsToDate:
    expr = build_formatted_time(exp.TsOrDsToDate)(args, dialect)
    expr.set("safe", True)
    return expr


def _build_named_struct(args: list) -> exp.Struct:
    """Map named_struct('k', v, ...) to exp.Struct so _annotate_struct sees it."""
    expressions: list[exp.Expression] = []
    for i in range(0, len(args) - 1, 2):
        key, value = args[i], args[i + 1]
        name = key.name
        expressions.append(exp.PropertyEQ(this=exp.to_identifier(name), expression=value))
    return exp.Struct(expressions=expressions)


def _build_date_add(args: list) -> exp.TsOrDsAdd:
    expression = seq_get(args, 1)
    if expression:
        expression = expression * -1

    return exp.TsOrDsAdd(
        this=seq_get(args, 0), expression=expression, unit=exp.Literal.string("DAY")
    )


class HiveParser(parser.Parser):
    LOG_DEFAULTS_TO_LN = True
    STRICT_CAST = False
    VALUES_FOLLOWED_BY_PAREN = False
    JOINS_HAVE_EQUAL_PRECEDENCE = True
    ADD_JOIN_ON_TRUE = True
    ALTER_TABLE_PARTITIONS = True

    CHANGE_COLUMN_ALTER_SYNTAX = False
    # Whether the dialect supports using ALTER COLUMN syntax with CHANGE COLUMN.

    FUNCTION_PARSERS = {
        **parser.Parser.FUNCTION_PARSERS,
        "PERCENTILE": lambda self: self._parse_distinct_arg_function(exp.Quantile),
        "PERCENTILE_APPROX": lambda self: self._parse_distinct_arg_function(exp.ApproxQuantile),
    }

    FUNCTIONS = {
        **parser.Parser.FUNCTIONS,
        "BASE64": exp.ToBase64.from_arg_list,
        "COLLECT_LIST": lambda args: exp.ArrayAgg(this=seq_get(args, 0), nulls_excluded=True),
        "COLLECT_SET": exp.ArrayUniqueAgg.from_arg_list,
        "DATE_ADD": lambda args: exp.TsOrDsAdd(
            this=seq_get(args, 0), expression=seq_get(args, 1), unit=exp.Literal.string("DAY")
        ),
        "DATE_FORMAT": lambda args, dialect: build_formatted_time(exp.TimeToStr)(
            [
                exp.TimeStrToTime(this=seq_get(args, 0)),
                seq_get(args, 1),
            ],
            dialect,
        ),
        "DATE_SUB": _build_date_add,
        "DATEDIFF": lambda args: exp.DateDiff(
            this=exp.TsOrDsToDate(this=seq_get(args, 0)),
            expression=exp.TsOrDsToDate(this=seq_get(args, 1)),
        ),
        "DAY": lambda args: exp.Day(this=exp.TsOrDsToDate(this=seq_get(args, 0))),
        "FIRST": build_with_ignore_nulls(exp.First),
        "FIRST_VALUE": build_with_ignore_nulls(exp.FirstValue),
        "FROM_UNIXTIME": build_formatted_time(exp.UnixToStr, default=True),
        "GET_JSON_OBJECT": lambda args, dialect: exp.JSONExtractScalar(
            this=seq_get(args, 0), expression=dialect.to_json_path(seq_get(args, 1))
        ),
        "LAST": build_with_ignore_nulls(exp.Last),
        "LAST_VALUE": build_with_ignore_nulls(exp.LastValue),
        "MAP": parser.build_var_map,
        "MONTH": lambda args: exp.Month(this=exp.TsOrDsToDate.from_arg_list(args)),
        "NAMED_STRUCT": _build_named_struct,
        "REGEXP_EXTRACT": build_regexp_extract(exp.RegexpExtract),
        "REGEXP_EXTRACT_ALL": build_regexp_extract(exp.RegexpExtractAll),
        "SEQUENCE": exp.GenerateSeries.from_arg_list,
        "SIZE": exp.ArraySize.from_arg_list,
        "SPLIT": exp.RegexpSplit.from_arg_list,
        "STR_TO_MAP": lambda args: exp.StrToMap(
            this=seq_get(args, 0),
            pair_delim=seq_get(args, 1) or exp.Literal.string(","),
            key_value_delim=seq_get(args, 2) or exp.Literal.string(":"),
        ),
        "TO_DATE": _build_to_date,
        "TO_JSON": exp.JSONFormat.from_arg_list,
        "TRUNC": exp.TimestampTrunc.from_arg_list,
        "UNBASE64": exp.FromBase64.from_arg_list,
        "UNIX_TIMESTAMP": lambda args, dialect: build_formatted_time(exp.StrToUnix, default=True)(
            args or [exp.CurrentTimestamp()], dialect
        ),
        "YEAR": lambda args: exp.Year(this=exp.TsOrDsToDate.from_arg_list(args)),
    }

    NO_PAREN_FUNCTION_PARSERS = {
        **parser.Parser.NO_PAREN_FUNCTION_PARSERS,
        "TRANSFORM": lambda self: self._parse_transform(),
    }

    NO_PAREN_FUNCTIONS = {
        k: v for k, v in parser.Parser.NO_PAREN_FUNCTIONS.items() if k != TokenType.CURRENT_TIME
    }

    PROPERTY_PARSERS = {
        **parser.Parser.PROPERTY_PARSERS,
        "SERDEPROPERTIES": lambda self: exp.SerdeProperties(
            expressions=self._parse_wrapped_csv(self._parse_property)
        ),
        "USING": lambda self: self._parse_using_property(),
    }

    ALTER_PARSERS = {
        **parser.Parser.ALTER_PARSERS,
        "CHANGE": lambda self: self._parse_alter_table_change(),
    }

    def _parse_transform(self) -> exp.Transform | exp.QueryTransform | None:
        if not self._match(TokenType.L_PAREN, advance=False):
            self._retreat(self._index - 1)
            return None

        args = self._parse_wrapped_csv(self._parse_lambda)
        row_format_before = self._parse_row_format(match_row=True)

        record_writer = None
        if self._match_text_seq("RECORDWRITER"):
            record_writer = self._parse_string()

        if not self._match(TokenType.USING):
            return exp.Transform.from_arg_list(args)

        command_script = self._parse_string()

        self._match(TokenType.ALIAS)
        schema = self._parse_schema()

        row_format_after = self._parse_row_format(match_row=True)
        record_reader = None
        if self._match_text_seq("RECORDREADER"):
            record_reader = self._parse_string()

        return self.expression(
            exp.QueryTransform(
                expressions=args,
                command_script=command_script,
                schema=schema,
                row_format_before=row_format_before,
                record_writer=record_writer,
                row_format_after=row_format_after,
                record_reader=record_reader,
            )
        )

    def _parse_types(
        self,
        check_func: bool = False,
        schema: bool = False,
        allow_identifiers: bool = True,
        with_collation: bool = False,
    ) -> exp.Expr | None:
        """
        Spark (and most likely Hive) treats casts to CHAR(length) and VARCHAR(length) as casts to
        STRING in all contexts except for schema definitions. For example, this is in Spark v3.4.0:

            spark-sql (default)> select cast(1234 as varchar(2));
            23/06/06 15:51:18 WARN CharVarcharUtils: The Spark cast operator does not support
            char/varchar type and simply treats them as string type. Please use string type
            directly to avoid confusion. Otherwise, you can set spark.sql.legacy.charVarcharAsString
            to true, so that Spark treat them as string type as same as Spark 3.0 and earlier

            1234
            Time taken: 4.265 seconds, Fetched 1 row(s)

        This shows that Spark doesn't truncate the value into '12', which is inconsistent with
        what other dialects (e.g. postgres) do, so we need to drop the length to transpile correctly.

        Reference: https://spark.apache.org/docs/latest/sql-ref-datatypes.html
        """
        this = super()._parse_types(
            check_func=check_func,
            schema=schema,
            allow_identifiers=allow_identifiers,
            with_collation=with_collation,
        )

        if this and not schema:

            def _to_text(node: exp.Expr) -> exp.Expr:
                if isinstance(node, exp.DataType) and node.is_type("char", "varchar"):
                    node.set("this", exp.DType.TEXT)
                    node.set("expressions", None)
                return node

            return this.transform(_to_text, copy=False)

        return this

    def _parse_alter_table_change(self) -> exp.Expr | None:
        self._match(TokenType.COLUMN)
        this = self._parse_field(any_token=True)

        if self.CHANGE_COLUMN_ALTER_SYNTAX and self._match_text_seq("TYPE"):
            return self.expression(exp.AlterColumn(this=this, dtype=self._parse_types(schema=True)))

        column_new = self._parse_field(any_token=True)
        dtype = self._parse_types(schema=True)

        comment = self._match(TokenType.COMMENT) and self._parse_string()

        if not this or not column_new or not dtype:
            self.raise_error(
                "Expected 'CHANGE COLUMN' to be followed by 'column_name' 'column_name' 'data_type'"
            )

        return self.expression(
            exp.AlterColumn(this=this, rename_to=column_new, dtype=dtype, comment=comment)
        )

    def _parse_using_property(self) -> exp.Property:
        if self._match_texts(("JAR", "FILE", "ARCHIVE")):
            kind = self._prev.text.upper()
            return exp.UsingProperty(this=self._parse_string(), kind=kind)

        return self._parse_property_assignment(exp.FileFormatProperty)

    def _parse_partition_and_order(
        self,
    ) -> tuple[list[exp.Expr], exp.Expr | None]:
        return (
            (
                self._parse_csv(self._parse_assignment)
                if self._match_set({TokenType.PARTITION_BY, TokenType.DISTRIBUTE_BY})
                else []
            ),
            super()._parse_order(skip_order_token=self._match(TokenType.SORT_BY)),
        )

    def _parse_parameter(self) -> exp.Parameter:
        self._match(TokenType.L_BRACE)
        this = self._parse_identifier() or self._parse_primary_or_var()
        expression = self._match(TokenType.COLON) and (
            self._parse_identifier() or self._parse_primary_or_var()
        )
        self._match(TokenType.R_BRACE)
        return self.expression(exp.Parameter(this=this, expression=expression))

    def _to_prop_eq(self, expression: exp.Expr, index: int) -> exp.Expr:
        if expression.is_star:
            return expression

        if isinstance(expression, exp.Column):
            key = expression.this
        else:
            key = exp.to_identifier(f"col{index + 1}")

        return self.expression(exp.PropertyEQ(this=key, expression=expression))



import unittest

from sqlglot.errors import highlight_sql, ANSI_UNDERLINE, ANSI_RESET


class TestErrors(unittest.TestCase):
    def test_highlight_sql_single_character(self):
        sql = "SELECT a FROM t"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(7, 7)])

        self.assertEqual(start_ctx, "SELECT ")
        self.assertEqual(highlight, "a")
        self.assertEqual(end_ctx, " FROM t")
        self.assertEqual(formatted, f"SELECT {ANSI_UNDERLINE}a{ANSI_RESET} FROM t")

    def test_highlight_sql_multi_character(self):
        sql = "SELECT foo FROM table"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(7, 9)])

        self.assertEqual(start_ctx, "SELECT ")
        self.assertEqual(highlight, "foo")
        self.assertEqual(end_ctx, " FROM table")
        self.assertEqual(formatted, f"SELECT {ANSI_UNDERLINE}foo{ANSI_RESET} FROM table")

    def test_highlight_sql_multiple_highlights(self):
        sql = "SELECT a, b, c FROM table"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(7, 7), (10, 10)])

        self.assertEqual(start_ctx, "SELECT ")
        self.assertEqual(highlight, "a, b")
        self.assertEqual(end_ctx, ", c FROM table")
        self.assertEqual(
            formatted,
            f"SELECT {ANSI_UNDERLINE}a{ANSI_RESET}, {ANSI_UNDERLINE}b{ANSI_RESET}, c FROM table",
        )

    def test_highlight_sql_at_end(self):
        sql = "SELECT a FROM t"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(14, 14)])

        self.assertEqual(start_ctx, "SELECT a FROM ")
        self.assertEqual(highlight, "t")
        self.assertEqual(end_ctx, "")
        self.assertEqual(formatted, f"SELECT a FROM {ANSI_UNDERLINE}t{ANSI_RESET}")

    def test_highlight_sql_entire_string(self):
        sql = "SELECT a"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(0, 7)])

        self.assertEqual(start_ctx, "")
        self.assertEqual(highlight, "SELECT a")
        self.assertEqual(end_ctx, "")
        self.assertEqual(formatted, f"{ANSI_UNDERLINE}SELECT a{ANSI_RESET}")

    def test_highlight_sql_adjacent_highlights(self):
        sql = "SELECT ab FROM t"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(7, 7), (8, 8)])

        self.assertEqual(start_ctx, "SELECT ")
        self.assertEqual(highlight, "ab")
        self.assertEqual(end_ctx, " FROM t")
        self.assertEqual(
            formatted, f"SELECT {ANSI_UNDERLINE}a{ANSI_RESET}{ANSI_UNDERLINE}b{ANSI_RESET} FROM t"
        )

    def test_highlight_sql_small_context_length(self):
        sql = "SELECT a, b, c FROM table WHERE x = 1"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(
            sql, [(7, 7), (10, 10)], context_length=5
        )

        self.assertEqual(start_ctx, "LECT ")
        self.assertEqual(highlight, "a, b")
        self.assertEqual(end_ctx, ", c F")
        self.assertEqual(
            formatted, f"LECT {ANSI_UNDERLINE}a{ANSI_RESET}, {ANSI_UNDERLINE}b{ANSI_RESET}, c F"
        )

    def test_highlight_sql_empty_positions(self):
        sql = "SELECT a FROM t"
        with self.assertRaises(ValueError):
            highlight_sql(sql, [])

    def test_highlight_sql_partial_overlap(self):
        sql = "SELECT foo FROM table"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(
            sql,
            [(7, 9), (8, 10)],  # "foo" and "oo "
        )

        self.assertEqual(start_ctx, "SELECT ")
        self.assertEqual(highlight, "foo ")
        self.assertEqual(end_ctx, "FROM table")
        self.assertEqual(
            formatted,
            f"SELECT {ANSI_UNDERLINE}foo{ANSI_RESET}{ANSI_UNDERLINE} {ANSI_RESET}FROM table",
        )

    def test_highlight_sql_full_overlap(self):
        sql = "SELECT foobar FROM table"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(
            sql,
            [(7, 12), (9, 11)],  # "foobar" and "oba"
        )

        self.assertEqual(start_ctx, "SELECT ")
        self.assertEqual(highlight, "foobar")
        self.assertEqual(end_ctx, " FROM table")
        self.assertEqual(formatted, f"SELECT {ANSI_UNDERLINE}foobar{ANSI_RESET} FROM table")

    def test_highlight_sql_identical_positions(self):
        sql = "SELECT a FROM t"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(7, 7), (7, 7)])

        self.assertEqual(start_ctx, "SELECT ")
        self.assertEqual(highlight, "a")
        self.assertEqual(end_ctx, " FROM t")
        self.assertEqual(formatted, f"SELECT {ANSI_UNDERLINE}a{ANSI_RESET} FROM t")

    def test_highlight_sql_reversed_positions(self):
        sql = "SELECT a, b FROM table"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(10, 10), (7, 7)])

        self.assertEqual(start_ctx, "SELECT ")
        self.assertEqual(highlight, "a, b")
        self.assertEqual(end_ctx, " FROM table")
        self.assertEqual(
            formatted,
            f"SELECT {ANSI_UNDERLINE}a{ANSI_RESET}, {ANSI_UNDERLINE}b{ANSI_RESET} FROM table",
        )

    def test_highlight_sql_zero_context_length(self):
        sql = "SELECT a, b FROM table"
        formatted, start_ctx, highlight, end_ctx = highlight_sql(sql, [(7, 7)], context_length=0)

        self.assertEqual(start_ctx, "")
        self.assertEqual(end_ctx, "")
        self.assertEqual(highlight, "a")
        self.assertEqual(formatted, f"{ANSI_UNDERLINE}a{ANSI_RESET}")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations


from sqlglot import exp, parser
from sqlglot.dialects.dialect import (
    binary_from_function,
    build_date_delta,
    build_formatted_time,
    build_timetostr_or_tochar,
    build_trunc,
)
from sqlglot.helper import seq_get
from sqlglot.tokens import TokenType

DATE_UNITS = {"DAY", "WEEK", "MONTH", "YEAR", "HOUR", "MINUTE", "SECOND"}


# https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/zeroifnull.htm
def _build_zeroifnull(args: list) -> exp.If:
    cond = exp.Is(this=seq_get(args, 0), expression=exp.Null())
    return exp.If(this=cond, true=exp.Literal.number(0), false=seq_get(args, 0))


# https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/nullifzero.htm
def _build_nullifzero(args: list) -> exp.If:
    cond = exp.EQ(this=seq_get(args, 0), expression=exp.Literal.number(0))
    return exp.If(this=cond, true=exp.Null(), false=seq_get(args, 0))


class ExasolParser(parser.Parser):
    TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS | {
        TokenType.ANTI,
        TokenType.SEMI,
    }

    FUNCTIONS = {
        **parser.Parser.FUNCTIONS,
        **{f"ADD_{unit}S": build_date_delta(exp.DateAdd, default_unit=unit) for unit in DATE_UNITS},
        **{
            f"{unit}S_BETWEEN": build_date_delta(exp.DateDiff, default_unit=unit)
            for unit in DATE_UNITS
        },
        "APPROXIMATE_COUNT_DISTINCT": exp.ApproxDistinct.from_arg_list,
        "BIT_AND": binary_from_function(exp.BitwiseAnd),
        "BIT_OR": binary_from_function(exp.BitwiseOr),
        "BIT_XOR": binary_from_function(exp.BitwiseXor),
        "BIT_NOT": lambda args: exp.BitwiseNot(this=seq_get(args, 0)),
        "BIT_LSHIFT": binary_from_function(exp.BitwiseLeftShift),
        "BIT_RSHIFT": binary_from_function(exp.BitwiseRightShift),
        # https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/convert_tz.htm
        "CONVERT_TZ": lambda args: exp.ConvertTimezone(
            source_tz=seq_get(args, 1),
            target_tz=seq_get(args, 2),
            timestamp=seq_get(args, 0),
            options=seq_get(args, 3),
        ),
        "CURDATE": exp.CurrentDate.from_arg_list,
        # https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/date_trunc.htm#DATE_TRUNC
        "DATE_TRUNC": lambda args: exp.TimestampTrunc(this=seq_get(args, 1), unit=seq_get(args, 0)),
        "DIV": binary_from_function(exp.IntDiv),
        "EVERY": lambda args: exp.All(this=seq_get(args, 0)),
        "EDIT_DISTANCE": exp.Levenshtein.from_arg_list,
        "FROM_POSIX_TIME": exp.UnixToTime.from_arg_list,
        "HASH_SHA": exp.SHA.from_arg_list,
        "HASH_SHA1": exp.SHA.from_arg_list,
        "HASH_MD5": exp.MD5.from_arg_list,
        "HASHTYPE_MD5": exp.MD5Digest.from_arg_list,
        "HASH_SHA256": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(256)),
        "HASH_SHA512": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(512)),
        "NOW": exp.CurrentTimestamp.from_arg_list,
        "NULLIFZERO": _build_nullifzero,
        "REGEXP_LIKE": lambda args: exp.RegexpLike(
            this=seq_get(args, 0),
            expression=seq_get(args, 1),
            flag=seq_get(args, 2),
            full_match=True,
        ),
        "REGEXP_SUBSTR": exp.RegexpExtract.from_arg_list,
        "REGEXP_REPLACE": lambda args: exp.RegexpReplace(
            this=seq_get(args, 0),
            expression=seq_get(args, 1),
            replacement=seq_get(args, 2),
            position=seq_get(args, 3),
            occurrence=seq_get(args, 4),
        ),
        "TRUNC": build_trunc,
        "TRUNCATE": build_trunc,
        "TO_CHAR": build_timetostr_or_tochar,
        "TO_DATE": build_formatted_time(exp.TsOrDsToDate),
        "USER": exp.CurrentUser.from_arg_list,
        "VAR_POP": exp.VariancePop.from_arg_list,
        "ZEROIFNULL": _build_zeroifnull,
    }
    CONSTRAINT_PARSERS = {
        **parser.Parser.CONSTRAINT_PARSERS,
        "COMMENT": lambda self: self.expression(
            exp.CommentColumnConstraint(this=self._match(TokenType.IS) and self._parse_string())
        ),
    }

    RANGE_PARSERS = {
        **parser.Parser.RANGE_PARSERS,
        TokenType.RLIKE: lambda self, this: self.expression(
            exp.RegexpLike(this=this, expression=self._parse_bitwise(), full_match=True)
        ),
    }

    FUNC_TOKENS = {
        *parser.Parser.FUNC_TOKENS,
        TokenType.SYSTIMESTAMP,
    }

    NO_PAREN_FUNCTIONS = {
        **parser.Parser.NO_PAREN_FUNCTIONS,
        TokenType.SYSTIMESTAMP: exp.Systimestamp,
        TokenType.CURRENT_SCHEMA: exp.CurrentSchema,
    }

    FUNCTION_PARSERS = {
        **parser.Parser.FUNCTION_PARSERS,
        # https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/group_concat.htm
        "GROUP_CONCAT": lambda self: self._parse_group_concat(),
        # https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/listagg.htm
        "LISTAGG": lambda self: self._parse_string_agg(),
        # https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/json_value.htm
        "JSON_VALUE": lambda self: self._parse_json_value(),
        # https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/json_extract.htm
        "JSON_EXTRACT": lambda self: self._parse_json_extract(),
    }

    def _parse_statement(self) -> exp.Expr | None:
        # https://docs.exasol.com/db/latest/sql/open_schema.htm
        if self._match_text_seq("OPEN", "SCHEMA"):
            return self.expression(exp.Use(this=self._parse_table(schema=False)))
        return super()._parse_statement()

    def _parse_column(self) -> exp.Expr | None:
        column = super()._parse_column()
        if not isinstance(column, exp.Column):
            return column
        table_ident = column.args.get("table")
        if (
            isinstance(table_ident, exp.Identifier)
            and table_ident.name.upper() == "LOCAL"
            and not bool(table_ident.args.get("quoted"))
        ):
            column.set("table", None)
        return column

    def _parse_json_extract(self) -> exp.JSONExtract:
        args = self._parse_expressions()
        self._match_r_paren()

        expression = exp.JSONExtract(expressions=args)

        if self._match_text_seq("EMITS"):
            expression.set("emits", self._parse_schema())

        return expression

    ODBC_DATETIME_LITERALS = {
        "d": exp.Date,
        "ts": exp.Timestamp,
    }



from sqlglot import TokenType
from sqlglot.dialects.mysql import MySQL
from sqlglot.generators.singlestore import SingleStoreGenerator
from sqlglot.parsers.singlestore import SingleStoreParser, cast_to_time6


class SingleStore(MySQL):
    SUPPORTS_ORDER_BY_ALL = True

    MYSQL_INVERSE_TIME_MAPPING = MySQL.INVERSE_TIME_MAPPING
    MYSQL_INVERSE_TIME_TRIE = MySQL.INVERSE_TIME_TRIE
    CAST_TO_TIME6 = staticmethod(cast_to_time6)

    TIME_MAPPING: dict[str, str] = {
        "D": "%u",  # Day of week (1-7)
        "DD": "%d",  # day of month (01-31)
        "DY": "%a",  # abbreviated name of day
        "HH": "%I",  # Hour of day (01-12)
        "HH12": "%I",  # alias for HH
        "HH24": "%H",  # Hour of day (00-23)
        "MI": "%M",  # Minute (00-59)
        "MM": "%m",  # Month (01-12; January = 01)
        "MON": "%b",  # Abbreviated name of month
        "MONTH": "%B",  # Name of month
        "SS": "%S",  # Second (00-59)
        "RR": "%y",  # 15
        "YY": "%y",  # 15
        "YYYY": "%Y",  # 2015
        "FF6": "%f",  # only 6 digits are supported in python formats
    }

    VECTOR_TYPE_ALIASES = {
        "I8": "TINYINT",
        "I16": "SMALLINT",
        "I32": "INT",
        "I64": "BIGINT",
        "F32": "FLOAT",
        "F64": "DOUBLE",
    }

    INVERSE_VECTOR_TYPE_ALIASES = {v: k for k, v in VECTOR_TYPE_ALIASES.items()}

    class Tokenizer(MySQL.Tokenizer):
        BYTE_STRINGS = [("e'", "'"), ("E'", "'")]

        KEYWORDS = {
            **MySQL.Tokenizer.KEYWORDS,
            "BSON": TokenType.JSONB,
            "GEOGRAPHYPOINT": TokenType.GEOGRAPHYPOINT,
            "TIMESTAMP": TokenType.TIMESTAMP,
            "UTC_DATE": TokenType.UTC_DATE,
            "UTC_TIME": TokenType.UTC_TIME,
            "UTC_TIMESTAMP": TokenType.UTC_TIMESTAMP,
            ":>": TokenType.COLON_GT,
            "!:>": TokenType.NCOLON_GT,
            "::$": TokenType.DCOLONDOLLAR,
            "::%": TokenType.DCOLONPERCENT,
            "::?": TokenType.DCOLONQMARK,
            "RECORD": TokenType.STRUCT,
        }

    Parser = SingleStoreParser

    Generator = SingleStoreGenerator
